# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import Command


class ArchiveCommand(Command):
    def __init__(self, p_todolist, p_archive_list):
        """
        Constructor.

        p_todolist where all completed items will be moved from.
        p_archive_list is the list where the items go to. This can be a
        TodoListBase class which does no dependency checking, so a better
        choice for huge done.txt files.
        """
        super().__init__([], p_todolist)
        self.archive = p_archive_list

    def execute(self):
        for todo in [t for t in self.todolist.todos() if t.is_completed()]:
            self.archive.add_todo(todo)
            self.todolist.delete(todo)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import Command, InvalidCommandArgument
from topydo.lib.Config import config
from topydo.lib.Sorter import Sorter


class SortCommand(Command):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

    def execute(self):
        if not super().execute():
            return False

        try:
            expression = self.argument(0)
        except InvalidCommandArgument:
            expression = config().sort_string()

        sorter = Sorter(expression)  # TODO: validate
        sorted_todos = sorter.sort(self.todolist.todos())

        self.todolist.replace(sorted_todos)

    def usage(self):
        return """Synopsis: sort [<EXPRESSION>]"""

    def help(self):
        return """\
Sorts the todo file according to the EXPRESSION. If no EXPRESSION is given, the
expression in the configuration is used.

The EXPRESSION is a comma separated list of attributes to sort on. The list is
evaluated in order, which means that the first attribute takes higher
precedence, then the second, etc.

The following sort attributes are supported:

* priority       - Sort by priority
* creation       - Sort by creation date
* completed      - Sort by completion state
* importance     - Sort by importance
* importance-avg - Sort by average importance (based on parent items)
* text           - Sort by text
* <TAG>          - Sort by values of the given TAG

Each item can optionally be prefixed with asc: and desc: to specify ascending
or descending sort respectively. If not specified, ascending sort is assumed.

Example:

    desc:importance,due,desc:priority\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module deals with relative dates (2d, 5y, Monday, today, etc.) """

import calendar
import re
from datetime import date, timedelta


def _add_months(p_sourcedate, p_months):
    """
    Adds a number of months to the source date.

    Takes into account shorter months and leap years and such.

    https://stackoverflow.com/questions/4130922/how-to-increment-datetime-month-in-python
    """
    month = p_sourcedate.month - 1 + p_months
    year = p_sourcedate.year + month // 12
    month = month % 12 + 1
    day = min(p_sourcedate.day, calendar.monthrange(year, month)[1])

    return date(year, month, day)


def _add_business_days(p_sourcedate, p_bdays):
    """ Adds a number of business days to the source date. """
    result = p_sourcedate
    delta = 1 if p_bdays > 0 else -1

    while abs(p_bdays) > 0:
        result += timedelta(delta)

        weekday = result.weekday()
        if weekday >= 5:
            continue

        p_bdays = p_bdays - 1 if delta > 0 else p_bdays + 1

    return result


def _convert_pattern(p_length, p_periodunit, p_offset=None):
    """
    Converts a pattern in the form [0-9][dwmyb] and returns a date from the
    offset with the period of time added to it.
    """
    result = None

    p_offset = p_offset or date.today()
    p_length = int(p_length)

    if p_periodunit == 'd':
        result = p_offset + timedelta(p_length)
    elif p_periodunit == 'w':
        result = p_offset + timedelta(weeks=p_length)
    elif p_periodunit == 'm':
        result = _add_months(p_offset, p_length)
    elif p_periodunit == 'y':
        result = _add_months(p_offset, p_length * 12)
    elif p_periodunit == 'b':
        result = _add_business_days(p_offset, p_length)

    return result


def _convert_weekday_pattern(p_weekday):
    """
    Converts a weekday name to an absolute date.

    When today's day of the week is entered, it will return next week's date.
    """
    day_value = {
        'mo': 0,
        'tu': 1,
        'we': 2,
        'th': 3,
        'fr': 4,
        'sa': 5,
        'su': 6
        }

    target_day_string = p_weekday[:2].lower()
    target_day = day_value[target_day_string]

    day = date.today().weekday()

    shift = 7 - (day - target_day) % 7
    return date.today() + timedelta(shift)


def relative_date_to_date(p_date, p_offset=None):
    """
    Transforms a relative date into a date object.

    The following formats are understood:

    * [0-9][dwmy]
    * 'yesterday', 'today' or 'tomorrow'
    * days of the week (in full or abbreviated)
    """
    result = None
    p_date = p_date.lower()
    p_offset = p_offset or date.today()

    relative = re.match('(?P<length>-?[0-9]+)(?P<period>[dwmyb])$',
                        p_date, re.I)

    monday = 'mo(n(day)?)?$'
    tuesday = 'tu(e(sday)?)?$'
    wednesday = 'we(d(nesday)?)?$'
    thursday = 'th(u(rsday)?)?$'
    friday = 'fr(i(day)?)?$'
    saturday = 'sa(t(urday)?)?$'
    sunday = 'su(n(day)?)?$'

    weekday = re.match('|'.join(
        [monday, tuesday, wednesday, thursday, friday, saturday, sunday]),
        p_date)

    if relative:
        length = relative.group('length')
        period = relative.group('period')
        result = _convert_pattern(length, period, p_offset)

    elif weekday:
        result = _convert_weekday_pattern(weekday.group(0))

    elif re.match('tod(ay)?$', p_date):
        result = _convert_pattern('0', 'd')

    elif re.match('tom(orrow)?$', p_date):
        result = _convert_pattern('1', 'd')

    elif re.match('yes(terday)?$', p_date):
        result = _convert_pattern('-1', 'd')

    return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.ExpressionCommand import ExpressionCommand
from topydo.lib.TodoListBase import InvalidTodoException


class MultiCommand(ExpressionCommand):
    """
    A common class for operations that can work with multiple todo IDs.
    """

    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.todos = []
        self.invalid_numbers = []
        self.is_expression = False
        self.multi_mode = True

    # pylint: disable=no-self-use
    def get_flags(self):
        """ Default implementation of getting specific flags. """
        return ("", [])

    def process_flag(self, p_option, p_value):
        """ Default implementation of processing specific flags. """
        raise NotImplementedError

    def _process_flags(self):
        opts, long_opts = self.get_flags()
        opts, args = self.getopt("xe" + opts, long_opts)

        for opt, value in opts:
            if opt == '-x':
                self.show_all = True
            elif opt == '-e':
                self.is_expression = True
            else:
                self.process_flag(opt, value)

        self.args = args

    def get_todos_from_expr(self):
        self.todos = self._view().todos

    def get_todos(self):
        """ Gets todo objects from supplied todo IDs. """
        if self.is_expression:
            self.get_todos_from_expr()
        else:
            if self.last_argument:
                numbers = self.args[:-1]
            else:
                numbers = self.args

            for number in numbers:
                try:
                    self.todos.append(self.todolist.todo(number))
                except InvalidTodoException:
                    self.invalid_numbers.append(number)

    def _catch_todo_errors(self):
        """
        Returns None or list of error messages depending on number of valid
        todo objects and number of invalid todo IDs.

        In case of multiple invalid todo IDs we generate separate error message
        for each one of them with information about supplied ID.
        """
        errors = []

        if len(self.invalid_numbers) > 1 or len(self.invalid_numbers) > 0 and len(self.todos) > 0:
            for number in self.invalid_numbers:
                errors.append(u"Invalid todo number given: {}.".format(number))
        elif len(self.invalid_numbers) == 1 and len(self.todos) == 0:
            errors.append("Invalid todo number given.")
        elif len(self.todos) == 0 and len(self.invalid_numbers) == 0:
            errors.append(self.usage())

        if len(errors) > 0:
            return errors
        else:
            return None

    def _execute_multi_specific(self):
        """
        Operations specific for particular command dealing with multiple todo
        IDs.
        """
        raise NotImplementedError

    def _execute_not_multi(self):
        """
        Some commands can do something else besides operating on multiple todo
        IDs. This method is a wrapper for those other operations.
        """
        raise NotImplementedError

    def execute(self):
        if not super().execute():
            return False

        self._process_flags()

        if not self.multi_mode:
            self._execute_not_multi()
        else:
            self.get_todos()
            todo_errors = self._catch_todo_errors()

            if not todo_errors:
                self._execute_multi_specific()
            else:
                for error in todo_errors:
                    self.error(error)

        return True



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Contains the class for a directed graph. """


class DirectedGraph(object):
    """
    Represents a simple directed graph, used for tracking todo
    dependencies. The nodes are very simple: just integers.
    """

    def __init__(self):
        self._edges = {}
        self._edge_numbers = {}

    def add_node(self, p_id):
        """ Adds a node to the graph. """
        if not self.has_node(p_id):
            self._edges[p_id] = set()

    def has_node(self, p_id):
        """ Returns true iff the graph has the given node. """
        return p_id in self._edges

    def add_edge(self, p_from, p_to, p_id=None):
        """
        Adds an edge to the graph. The nodes will be added if they don't exist.

        The p_id is the id of the edge, if the client wishes to maintain this.
        """
        if not self.has_edge(p_from, p_to):
            if not self.has_node(p_from):
                self.add_node(p_from)

            if not self.has_node(p_to):
                self.add_node(p_to)

            self._edges[p_from].add(p_to)
            self._edge_numbers[(p_from, p_to)] = p_id

    def has_path(self, p_from, p_to):
        """
        Returns true iff there is a path from the first node to the second.
        """
        return p_to in self.reachable_nodes(p_from)

    def incoming_neighbors(self, p_id, p_recursive=False):
        """
        Returns a set of the direct neighbors that can reach the given node.
        """
        return self.reachable_nodes_reverse(p_id, p_recursive)

    def outgoing_neighbors(self, p_id, p_recursive=False):
        """
        Returns the set of the direct neighbors that the given node can reach.
        """
        return self.reachable_nodes(p_id, p_recursive)

    def reachable_nodes(self, p_id, p_recursive=True, p_reverse=False):
        """
        Returns the set of all neighbors that the given node can reach.

        If recursive, it will also return the neighbor's neighbors, etc.
        If reverse, the arrows are reversed and then the reachable neighbors
        are located.
        """
        stack = [p_id]
        visited = set()
        result = set()

        while len(stack):
            current = stack.pop()

            if current in visited or current not in self._edges:
                continue

            visited.add(current)

            if p_reverse:
                parents = [node for node, neighbors in self._edges.items()
                           if current in neighbors]

                stack = stack + parents
                result = result.union(parents)
            else:
                stack = stack + list(self._edges[current])
                result = result.union(self._edges[current])

            if not p_recursive:
                break

        return result

    def reachable_nodes_reverse(self, p_id, p_recursive=True):
        """ Find neighbors in the inverse graph. """
        return self.reachable_nodes(p_id, p_recursive, True)

    def remove_node(self, p_id, remove_unconnected_nodes=True):
        """ Removes a node from the graph. """
        if self.has_node(p_id):
            for neighbor in self.incoming_neighbors(p_id):
                self._edges[neighbor].remove(p_id)

            neighbors = set()
            if remove_unconnected_nodes:
                neighbors = self.outgoing_neighbors(p_id)

            del self._edges[p_id]

            for neighbor in neighbors:
                if self.is_isolated(neighbor):
                    self.remove_node(neighbor)

    def is_isolated(self, p_id):
        """
        Returns True iff the given node has no incoming or outgoing edges.
        """
        return(len(self.incoming_neighbors(p_id)) == 0
               and len(self.outgoing_neighbors(p_id)) == 0)

    def has_edge(self, p_from, p_to):
        """ Returns True when the graph has the given edge. """
        return p_from in self._edges and p_to in self._edges[p_from]

    def has_edge_id(self, p_id):
        """
        Returns True if the client registered an edge with the given id.
        """
        return p_id in self._edge_numbers.values()

    def edge_id(self, p_from, p_to):
        """
        Returns the edge ID given the from and to nodes.

        Returns None if the edge does not exist or has no value assigned.
        """
        try:
            return self._edge_numbers[(p_from, p_to)]
        except KeyError:
            return None

    def remove_edge(self, p_from, p_to, p_remove_unconnected_nodes=True):
        """
        Removes an edge from the graph.

        When remove_unconnected_nodes is True, then the nodes are also removed
        if they become isolated.
        """
        if self.has_edge(p_from, p_to):
            self._edges[p_from].remove(p_to)

        try:
            del self._edge_numbers[(p_from, p_to)]
        except KeyError:
            return None

        if p_remove_unconnected_nodes:
            if self.is_isolated(p_from):
                self.remove_node(p_from)

            if self.is_isolated(p_to):
                self.remove_node(p_to)

    def transitively_reduce(self):
        """
        Performs a transitive reduction on the graph.
        """
        removals = set()

        for from_node, neighbors in self._edges.items():
            childpairs = \
                [(c1, c2) for c1 in neighbors for c2 in neighbors if c1 != c2]

            for child1, child2 in childpairs:
                if self.has_path(child1, child2) \
                   and not self.has_path(child1, from_node):
                    removals.add((from_node, child2))

        for edge in removals:
            self.remove_edge(edge[0], edge[1])

    def dot(self, p_print_labels=True):
        """ Prints the graph in Dot format. """
        out = 'digraph g {\n'

        for from_node, neighbors in sorted(self._edges.items()):
            out += "  {}\n".format(from_node)

            for neighbor in sorted(neighbors):
                out += "  {} -> {}".format(from_node, neighbor)

                edge_id = self.edge_id(from_node, neighbor)
                if edge_id and p_print_labels:
                    out += ' [label="{}"]'.format(edge_id)

                out += "\n"

        out += '}\n'
        return out

    def __str__(self):
        """ Returns the graph in Dot format. """
        return self.dot()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2016 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module provides TopydoString to embed colors in a string. """

import collections


class TopydoString(collections.UserString):
    """
    Represents a string that also contains color information. A combination of
    (position, color) is maintained, where the position is the start position
    where a certain color should start.
    """

    def __init__(self, p_content, p_metadata=None):
        if isinstance(p_content, TopydoString):
            # don't nest topydostrings
            self.colors = p_content.colors
            self.metadata = p_content.metadata
            super().__init__(p_content.data)
        else:
            self.colors = {}
            super().__init__(p_content)

            # allows clients to pass arbitrary data with this string (e.g. a Todo
            # object)
            self.metadata = p_metadata

    def append(self, p_string, p_color):
        """
        Append a string with the given color (normal Color or an
        AbstractColor).
        """
        self.colors[len(self.data)] = p_color
        self.data += p_string

    def set_color(self, p_pos, p_color):
        """ Start using a color at the given position. """
        self.colors[p_pos] = p_color



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys

from topydo.lib.Config import config
from topydo.lib.ExpressionCommand import ExpressionCommand
from topydo.lib.Filter import HiddenTagFilter, InstanceFilter
from topydo.lib.ListFormat import ListFormatError
from topydo.lib.prettyprinters.Format import PrettyPrinterFormatFilter
from topydo.lib.printers.PrettyPrinter import pretty_printer_factory
from topydo.lib.Sorter import Sorter
from topydo.lib.TodoListBase import InvalidTodoException
from topydo.lib.Utils import get_terminal_size
from topydo.lib.View import View


class ListCommand(ExpressionCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.printer = None
        self.sort_expression = config().sort_string()
        self.group_expression = config().group_string()
        self.show_all = False
        self.ids = None
        self.format = config().list_format()

    def _poke_icalendar(self):
        """
        Attempts to import the icalendar package. Returns True if it
        succeeds, otherwise False.
        """
        try:
            import icalendar as _
        except ImportError:  # pragma: no cover
            self.error("icalendar package is not installed.")
            return False

        return True

    def _process_flags(self):
        opts, args = self.getopt('f:F:g:i:n:Ns:x')

        for opt, value in opts:
            if opt == '-x':
                self.show_all = True
            elif opt == '-s':
                self.sort_expression = value
            elif opt == '-f':
                if value == 'json':
                    from topydo.lib.printers.Json import JsonPrinter
                    self.printer = JsonPrinter()
                elif value == 'ical':
                    if self._poke_icalendar():
                        from topydo.lib.printers.Ical import IcalPrinter
                        self.printer = IcalPrinter(self.todolist)
                elif value == 'dot':
                    from topydo.lib.printers.Dot import DotPrinter
                    self.printer = DotPrinter(self.todolist)

                    # a graph without dependencies is not so useful, hence
                    # show all
                    self.show_all = True
                else:
                    self.printer = None
            elif opt == '-F':
                self.format = value
            elif opt == '-g':
                self.group_expression = value
            elif opt == '-N':
                # 2 lines are assumed to be taken up by printing the next prompt
                # display at least one item
                self.limit = ListCommand._N_lines()
            elif opt == '-n':
                try:
                    self.limit = int(value)
                except ValueError:
                    pass  # use default value in configuration
            elif opt == '-i':
                self.ids = value.split(',')

                # when a user requests a specific ID, it should always be shown
                self.show_all = True

        self.args = args

    def _filters(self):
        """
        Additional filters to:
            - select particular todo items given with the -i flag,
            - hide appropriately tagged items in the absense of the -x flag.
        """
        filters = super()._filters()

        if self.ids:
            def get_todo(p_id):
                """
                Safely obtains a todo item given the user-supplied ID.
                Returns None if an invalid ID was entered.
                """
                try:
                    return self.todolist.todo(p_id)
                except InvalidTodoException:
                    return None

            todos = [get_todo(i) for i in self.ids]
            filters.append(InstanceFilter(todos))

        if not self.show_all:
            filters.append(HiddenTagFilter())

        return filters

    def _print(self):
        """
        Prints the todos in the right format.

        Defaults to normal text output (with possible colors and other pretty
        printing). If a format was specified on the commandline, this format is
        sent to the output.
        """
        if self.printer is None:
            # create a standard printer with some filters
            indent = config().list_indent()
            final_format = ' ' * indent + self.format

            filters = []
            filters.append(PrettyPrinterFormatFilter(self.todolist, final_format))

            self.printer = pretty_printer_factory(self.todolist, filters)

        try:
            if self.group_expression:
                self.out(self.printer.print_groups(self._view().groups))
            else:
                self.out(self.printer.print_list(self._view().todos))
        except ListFormatError:
            self.error('Error while parsing format string (list_format config'
                       ' option or -F)')

    def _view(self):
        sorter = Sorter(self.sort_expression, self.group_expression)
        filters = self._filters()

        return View(sorter, filters, self.todolist)

    @staticmethod
    def _N_lines():
        ''' Determine how many lines to print, such that the number of items
            displayed will fit on the terminal (i.e one 'screen-ful' of items)

            This looks at the environmental prompt variable, and tries to determine
            how many lines it takes up.

            On Windows, it does this by looking for the '$_' sequence, which indicates
            a new line, in the environmental variable PROMPT.

            Otherwise, it looks for a newline ('\n') in the environmental variable
            PS1.
        '''
        lines_in_prompt = 1     # prompt is assumed to take up one line, even
                                #   without any newlines in it
        if "win32" in sys.platform:
            lines_in_prompt += 1  # Windows will typically print a free line after
                                  #   the program output
            a = re.findall(r'\$_', os.getenv('PROMPT', ''))
            lines_in_prompt += len(a)
        else:
            a = re.findall('\\n', os.getenv('PS1', ''))
            lines_in_prompt += len(a)
        n_lines = get_terminal_size().lines - lines_in_prompt

        # print a minimum of one item
        n_lines = max(n_lines, 1)

        return n_lines

    def execute(self):
        if not super().execute():
            return False

        try:
            self._process_flags()
        except SyntaxError:  # pragma: no cover
            # importing icalendar failed, most likely due to Python 3.2
            self.error("icalendar is not supported in this Python version.")
            return False

        self._print()
        return True

    def usage(self):
        return """Synopsis: ls [-x] [-s <SORT EXPRESSION>]
[-g <GROUP EXPRESSION>] [-f <OUTPUT FORMAT>] [-F <FORMAT STRING>]
[-i <NUMBER 1>[,<NUMBER 2> ...]] [-N | -n <INTEGER>] [EXPRESSION]"""

    def help(self):
        return r"""\
Lists all relevant todos. A todo is relevant when:

* has not been completed yet,
* the start date (if present) has passed, and
* there are no subitems that need to be completed.

When an EXPRESSION is given, only the todos matching that EXPRESSION are shown.

-f : Specify the OUTPUT format, being 'text' (default), 'dot' or 'ical' or
     'json'.

     * 'text' - Text output with colors and indentation if applicable.
     * 'dot'  - Prints a dependency graph for the selected items in GraphViz
                Dot format.
     * 'ical' - iCalendar (RFC 2445). Is not supported in Python 3.2. Be aware
                that this is not a read-only operation, todo items may obtain
                an 'ical' tag with a unique ID. Completed todo items may be
                archived.
     * 'json' - Javascript Object Notation (JSON)

-F : Specify the format of the text ('text' format), which may contain
     placeholders that may be expanded if the todo has such attribute. If such
     attribute does not exist, then it expands to an empty string.

         %c: Absolute creation date.
         %C: Relative creation date.
         %d: Absolute due date.
         %D: Relative due date.
         %h: Relative due and start date (due in 3 days, started 3 days ago)
         %H: Like %h with creation date.
         %i: Todo number.
         %I: Todo number padded with spaces (always 3 characters wide).
         %k: List of tags separated by spaces (excluding hidden tags).
         %K: List of all tags separated by spaces.
         %l: Line number.
         %L: Line number padded with spaces (always 3 characters wide).
         %p: Priority.
         %P: Priority or placeholder space if no priority.
         %s: Todo text.
         %S: Todo text, truncated such that an item fits on one line.
         %t: Absolute start date.
         %T: Relative start date.
         %u: Todo's text-based ID.
         %U: Todo's text-based ID padded with spaces.
         %x: 'x' followed by absolute completion date.
         %X: 'x' followed by relative completion date.
         \%: Literal percent sign.

     Conditional characters can be added with blocks surrounded by curly
     braces, they will only appear when a placeholder expanded to a value.

     E.g. %{(}p{)} will print '(C)' when the todo item has priority C, or ''
     (empty string) when an item has no priority set.

     A tab character serves as a marker to start right alignment.
-g : Group items according to a GROUP EXPRESSION. A group expression is similar
     to a sort expression. Defaults to the group expression in the
     configuration.
-i : Comma separated list of todo IDs to print.
-n : Number of items to display. Defaults to the value in the configuration.
-N : Limit number of items displayed such that they fit on the terminal.
-s : Sort the list according to a SORT EXPRESSION. Defaults to the sort
     expression in the configuration.
-x : Show all todos (i.e. do not filter on dependencies, relevance, or hidden
     status).\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module is aware of all supported submodules and hands out a Command
instance based on an argument list.
"""

import sys

from topydo.lib.Config import ConfigError, config

SUBCOMMAND_MAP = {
    'add': 'AddCommand',
    'app': 'AppendCommand',
    'append': 'AppendCommand',
    'del': 'DeleteCommand',
    'dep': 'DepCommand',
    'depri': 'DepriCommand',
    'do': 'DoCommand',
    'edit': 'EditCommand',
    'exit': 'ExitCommand',  # used for the prompt
    'ls': 'ListCommand',
    'lscon': 'ListContextCommand',
    'lsprj': 'ListProjectCommand',
    'mod': 'ModifyCommand',
    'modify': 'ModifyCommand',
    'postpone': 'PostponeCommand',
    'pri': 'PriorityCommand',
    'quit': 'ExitCommand',
    'revert': 'RevertCommand',
    'rm': 'DeleteCommand',
    'sort': 'SortCommand',
    'tag': 'TagCommand',
}

def get_subcommand(p_args):
    """
    Retrieves the to-be executed Command and returns a tuple (Command, args).

    If args is an empty list, then the Command that corresponds with the
    default command specified in the configuration will be returned.

    If the first argument is 'help' and the second a valid subcommand, the
    help text this function returns the Command instance of that subcommand
    with a single argument 'help' (every Command has a help text).

    If no valid command could be found, the subcommand part of the tuple
    is None.
    """
    def import_subcommand(p_subcommand):
        """
        Returns the class of the requested subcommand. An invalid p_subcommand
        will result in an ImportError, since this is a programming mistake
        (most likely an error in the SUBCOMMAND_MAP).
        """
        classname = SUBCOMMAND_MAP[p_subcommand]
        modulename = 'topydo.commands.{}'.format(classname)

        __import__(modulename, globals(), locals(), [classname], 0)
        return getattr(sys.modules[modulename], classname)

    def join_args(p_cli_args, p_alias_args):
        """
        Returns properly joined args from alias definition and from user input.
        """
        if '{}' in p_alias_args:
            pos = p_alias_args.index('{}')
            args = p_alias_args[:pos] + p_cli_args + p_alias_args[pos+1:]
        else:
            args = p_alias_args + p_cli_args

        return args

    def resolve_alias(p_alias, p_args):
        """
        Resolves a subcommand alias and returns a tuple (Command, args).

        If alias resolves to non-existent command, main help message is
        returned.
        """
        try:
            real_subcommand, alias_args = alias_map[p_alias]
        except ValueError as ve:
            raise ConfigError(alias_map[p_alias]) from ve

        try:
            result = import_subcommand(real_subcommand)
            args = join_args(p_args, alias_args)
            return (result, args)
        except KeyError:
            return get_subcommand(['help'])

    result = None
    args = p_args
    alias_map = config().aliases()

    try:
        subcommand = p_args[0]

        if subcommand in alias_map:
            result, args = resolve_alias(subcommand, args[1:])
        elif subcommand in SUBCOMMAND_MAP:
            result = import_subcommand(subcommand)
            args = args[1:]
        elif subcommand == 'help':
            try:
                subcommand = args[1]

                if subcommand in SUBCOMMAND_MAP:
                    args = [subcommand, 'help']
                    return get_subcommand(args)
            except IndexError:
                # will result in empty result
                pass
        else:
            p_command = config().default_command()
            if p_command in alias_map:
                result, args = resolve_alias(p_command, args)
            elif p_command in SUBCOMMAND_MAP:
                result = import_subcommand(p_command)
                # leave args unchanged
    except IndexError:
        p_command = config().default_command()
        if p_command in alias_map:
            result, args = resolve_alias(p_command, args)
        elif p_command in SUBCOMMAND_MAP:
            result = import_subcommand(p_command)

    return (result, args)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import getopt

from topydo.lib.printers.PrettyPrinter import PrettyPrinter


class InvalidCommandArgument(Exception):
    pass


class Command(object):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        """
        Sets up the basic properties for executing a subcommand.

        p_args is a list of arguments that can be passed to this subcommand.
        These can be retrieved with argument().

        p_todolist is a reference to the todolist instance to operate on.

        p_out is a function to be called to print (standard) output. Defaults
        to a noop.

        p_err is a function to be called to print errors. Defaults to a noop.

        p_prompt is a function that accepts a prompt string as its own argument
        and returns the answer to that prompt (normally entered by the user in
        some way). The default is a noop prompt.
        """
        self.args = p_args
        self.todolist = p_todolist

        # inputs and outputs
        self.out = p_out
        self.error = p_err
        self.prompt = p_prompt

        # make pretty printer available
        self.printer = PrettyPrinter()

    def execute(self):
        """
        Execute the command. Intercepts the help subsubcommand to show the help
        text.
        """
        if self.args and self.argument(0) == "help":
            self.error(self.usage() + "\n\n" + self.help())
            return False

        return True

    def argument(self, p_number):
        """ Retrieves a value from the argument list at the given position. """
        try:
            return self.args[p_number]
        except IndexError as ie:
            raise InvalidCommandArgument from ie

    def getopt(self, p_flags, p_long=None):
        p_long = p_long or []

        try:
            result = getopt.getopt(self.args, p_flags, p_long)
        except getopt.GetoptError as goe:
            self.error(str(goe))
            result = ([], self.args)

        return result

    @classmethod
    def name(cls):
        """" Returns short-name of the command. """
        return cls.__name__[:-7].lower()  # strip 'Command'

    def execute_post_archive_actions(self):
        """ Runs various hooks that should take place after archiving. """
        pass

    def usage(self):
        """ Returns a one-line synopsis for this command. """
        raise NotImplementedError

    def help(self):
        """ Returns the help text for this command. """
        raise NotImplementedError



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import os
import shlex
import tempfile
from subprocess import CalledProcessError, check_call

from topydo.lib.Config import config
from topydo.lib.MultiCommand import MultiCommand
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.Todo import Todo


def _get_file_mtime(p_file):
    return os.stat(p_file.name).st_mtime

def _is_edited(p_orig_mtime, p_file):
    return p_orig_mtime < _get_file_mtime(p_file)

class EditCommand(MultiCommand):
    def __init__(self, p_args, p_todolist, p_output, p_error, p_input):
        super().__init__(p_args, p_todolist, p_output,
                                          p_error, p_input)

        self.editor = config().editor()
        self.is_expression = False
        self.edit_archive = False
        self.last_argument = False

    def get_flags(self):
        return ("dE:", [])

    def process_flag(self, p_option, p_value):
        if p_option == '-d':
            self.edit_archive = True
            self.multi_mode = False
        elif p_option == '-E':
            self.editor = shlex.split(p_value)

    def _process_flags(self):
        """
        Override to add an additional check after processing the flags: when
        there are no flags left after argument parsing, then it means we'll be
        editing the whole todo.txt file as a whole and therefore we're not in
        multi mode.
        """
        super()._process_flags()

        if len(self.args) == 0:
            self.multi_mode = False

    def _todos_to_temp(self):
        f = tempfile.NamedTemporaryFile(delete=False, suffix='.todo.txt')
        for todo in self.todos:
            f.write((todo.source() + "\n").encode('utf-8'))
        f.close()

        return f

    @staticmethod
    def _todos_from_temp(p_temp_file):
        with codecs.open(p_temp_file.name, encoding='utf-8') as temp:
            todos = temp.read().splitlines()

        todo_objs = []
        for todo in todos:
            todo_objs.append(Todo(todo))

        return todo_objs

    def _open_in_editor(self, p_file):
        try:
            return check_call(self.editor + [p_file])
        except CalledProcessError:
            self.error('Something went wrong in the editor...')
            return 1
        except OSError:
            self.error('There is no such editor as: ' + self.editor + '. '
                       'Check your configuration file, $TOPYDO_EDITOR, $EDITOR and/or $PATH')

    def _catch_todo_errors(self):
        errors = []

        if len(self.invalid_numbers) > 1 or len(self.invalid_numbers) > 0 and len(self.todos) > 0:
            for number in self.invalid_numbers:
                errors.append(u"Invalid todo number given: {}.".format(number))
        elif len(self.invalid_numbers) == 1 and len(self.todos) == 0:
            errors.append("Invalid todo number given.")

        if len(errors) > 0:
            return errors
        else:
            return None

    def _execute_multi_specific(self):
        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))

        temp_todos = self._todos_to_temp()
        orig_mtime = _get_file_mtime(temp_todos)

        if not self._open_in_editor(temp_todos.name):
            new_todos = EditCommand._todos_from_temp(temp_todos)

            if _is_edited(orig_mtime, temp_todos):
                modified = list(zip(self.todos, new_todos))
                for (todo, new_todo) in modified:
                    self.todolist.modify_todo(todo, new_todo.src)
                    self.out(self.printer.print_todo(todo))

                for todo in self.todos[len(modified):]:
                    self.todolist.delete(todo, p_leave_tags=True)

                for todo in new_todos[len(modified):]:
                    self.todolist.add_todo(todo)
                    self.out(self.printer.print_todo(todo))
            else:
                self.error('Editing aborted. Nothing to do.')
        else:
            self.error(self.usage())

        os.unlink(temp_todos.name)

    def _execute_not_multi(self):
        if self.edit_archive:
            archive = config().archive()

            return self._open_in_editor(archive) == 0
        else:
            todo = config().todotxt()

            return self._open_in_editor(todo) == 0


    def usage(self):
        return """Synopsis:
  edit [-E <EDITOR>]
  edit [-E <EDITOR>] <NUMBER 1> [<NUMBER 2> ...]
  edit [-E <EDITOR>] -e [-x] [EXPRESSION]
  edit [-E <EDITOR>] -d"""

    def help(self):
        return """\
Launches a text editor to edit todos.

Without any arguments it will just open the todo.txt file. Alternatively it can
edit todo item(s) with the given NUMBER(s) or edit relevant todos matching
the given EXPRESSION. See `topydo help ls` for more information on relevant
todo items. It is also possible to open the archive file.

The editor is chosen as follows:
    1. Check whether the -E flag is given and use it;
    2. Use the value of $TOPYDO_EDITOR in the environment;
    3. Use the value in the configuration file;
    4. Use the value of $EDITOR in the environment;
    5. If all else fails, use 'vi'.

-e : Treat the subsequent arguments as an EXPRESSION.
-E : Editor to start.
-x : Edit *all* todos matching the EXPRESSION (i.e. do not filter on
     dependencies or relevance).
-d : Open the archive file.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module deals with recurring tasks. """

from datetime import date, timedelta

from topydo.lib.Config import config
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.Todo import Todo


class NoRecurrenceException(Exception):
    pass


def advance_recurring_todo(p_todo, p_offset=None, p_strict=False):
    """
    Given a Todo item, return a new instance of a Todo item with the dates
    shifted according to the recurrence rule.

    Strict means that the real due date is taken as a offset, not today or a
    future date to determine the offset.

    When the todo item has no due date, then the date is used passed by the
    caller (defaulting to today).

    When no recurrence tag is present, an exception is raised.
    """
    todo = Todo(p_todo.source())
    pattern = todo.tag_value('rec')

    if not pattern:
        raise NoRecurrenceException()
    elif pattern.startswith('+'):
        p_strict = True
        # strip off the +
        pattern = pattern[1:]

    if p_strict:
        offset = p_todo.due_date() or p_offset or date.today()
    else:
        offset = p_offset or date.today()

    length = todo.length()
    new_due = relative_date_to_date(pattern, offset)

    if not new_due:
        raise NoRecurrenceException()

    if (not todo.start_date()) or todo.due_date():
    # pylint: disable=E1103
       todo.set_tag(config().tag_due(), new_due.isoformat())
       if todo.start_date():
           new_start = new_due - timedelta(length)
           todo.set_tag(config().tag_start(), new_start.isoformat())
    else: #only start date
        if p_strict:
            offset = todo.start_date()
        else:
            offset = p_offset or date.today()
        new_start = relative_date_to_date(pattern, offset)
        todo.set_tag(config().tag_start(), new_start.isoformat())

    if config().auto_creation_date():
        todo.set_creation_date(date.today())

    return todo



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from topydo.lib.Config import config
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.Utils import date_string_to_date


class Filter(object):
    def filter(self, p_todos):
        """
        Filters a list of todos. Truncates the list after p_limit todo
        items (or no maximum limit if omitted).
        """
        return [t for t in p_todos if self.match(t)]

    def match(self, _):
        raise NotImplementedError

    @property
    def order(self):
        return 50


class NegationFilter(Filter):
    def __init__(self, p_filter):
        self._filter = p_filter

    def match(self, p_todo):
        return not self._filter.match(p_todo)


class AndFilter(Filter):
    def __init__(self, p_filter1, p_filter2):
        self._filter1 = p_filter1
        self._filter2 = p_filter2

    def match(self, p_todo):
        return self._filter1.match(p_todo) and self._filter2.match(p_todo)


class OrFilter(Filter):
    def __init__(self, p_filter1, p_filter2):
        self._filter1 = p_filter1
        self._filter2 = p_filter2

    def match(self, p_todo):
        return self._filter1.match(p_todo) or self._filter2.match(p_todo)


class GrepFilter(Filter):
    """ Matches when the todo text contains a text. """

    def __init__(self, p_expression, p_case_sensitive=None):
        super().__init__()

        # convert to string in case we receive integers
        self.expression = p_expression

        if p_case_sensitive is not None:
            self.case_sensitive = p_case_sensitive
        else:
            # only be case sensitive when the expression contains at least one
            # capital character.
            self.case_sensitive = \
                len([c for c in self.expression if c.isupper()]) > 0

    def match(self, p_todo):
        expr = self.expression
        string = p_todo.source()
        if not self.case_sensitive:
            expr = expr.lower()
            string = string.lower()

        return string.find(expr) != -1


class RelevanceFilter(Filter):
    """
    Matches when the todo is relevant, i.e.:

    The item has not been completed AND
    The start date is blank, today or in the past, AND
    The priority is 'A', or the priority is 'B' with due date within 30 days, or
    the priority is 'C' with due date within 14 days.
    """

    def match(self, p_todo):
        active = p_todo.is_active()

        is_due = active
        is_due = is_due or p_todo.due_date() == None
        is_due = is_due or p_todo.priority() == 'A'
        is_due = is_due or (p_todo.priority() == 'B' and p_todo.days_till_due() <= 30)
        is_due = is_due or (p_todo.priority() == 'C' and p_todo.days_till_due() <= 14)

        return active and is_due

    @property
    def order(self):
        """
        Perform just after the DependencyFilter, but before any other filters.
        """
        return 20


class DependencyFilter(Filter):
    """ Matches when a todo has no unfinished child tasks.  """

    def __init__(self, p_todolist):
        """
        Constructor.

        Pass on a TodoList instance such that the dependencies can be
        looked up.
        """
        super().__init__()
        self.todolist = p_todolist

    def match(self, p_todo):
        """
        Returns True when there are no children that are uncompleted yet.
        """
        children = self.todolist.children(p_todo)
        uncompleted = [todo for todo in children if not todo.is_completed()]

        return not uncompleted

    @property
    def order(self):
        """
        Perform early before any other items are filtered out, possibly
        breaking the dependency chain.
        """
        return 10


class InstanceFilter(Filter):
    def __init__(self, p_todos):
        """
        Constructor.

        A filter which selects a number of Todo instances from a TodoList
        instance.

        This is handy for constructing a view given a plain list of Todo items.
        """
        super().__init__()
        self.todos = p_todos

    def match(self, p_todo):
        """
        Returns True when p_todo appears in the list of given todos.
        """
        try:
            self.todos.index(p_todo)
            return True
        except ValueError:
            return False


class HiddenTagFilter(Filter):
    def __init__(self):
        """
        Constructor.

        A filter which hides items marked with the approriate tag.
        """
        super().__init__()

    def match(self, p_todo):
        """
        Returns True when p_todo doesn't have a tag to mark it as hidden.
        """
        for my_tag in config().hidden_item_tags():
            my_values = p_todo.tag_values(my_tag)
            for my_value in my_values:
                if not my_value in (0, '0', False, 'False'):
                    return False

        return True


class LimitFilter(Filter):
    def __init__(self, p_limit):
        super().__init__()
        self.limit = p_limit

    def filter(self, p_todos):
        return p_todos[:self.limit] if self.limit >= 0 else p_todos

    @property
    def order(self):
        # should be performed at the very last step
        return 100

_OPERATOR_MATCH = r"(?P<operator><=?|=|>=?|!)?"


class OrdinalFilter(Filter):
    """ Base class for ordinal filters. """

    def __init__(self, p_expression, p_pattern):
        super().__init__()

        self.expression = p_expression

        match = re.match(p_pattern, self.expression)
        if match:
            try:
                self.key = match.group('key')
            except IndexError:
                pass
            self.operator = match.group('operator') or '='
            self.value = match.group('value')

    def compare_operands(self, p_operand1, p_operand2):
        """
        Returns True if conditional constructed from both operands and
        self.operator is valid. Returns False otherwise.
        """
        if self.operator == '<':
            return p_operand1 < p_operand2
        elif self.operator == '<=':
            return p_operand1 <= p_operand2
        elif self.operator == '=':
            return p_operand1 == p_operand2
        elif self.operator == '>=':
            return p_operand1 >= p_operand2
        elif self.operator == '>':
            return p_operand1 > p_operand2
        elif self.operator == '!':
            return p_operand1 != p_operand2

        return False

_VALUE_MATCH = r"(?P<value>\S+)"
_ORDINAL_TAG_MATCH = r"(?P<key>[^:]*):" + _OPERATOR_MATCH + _VALUE_MATCH


class OrdinalTagFilter(OrdinalFilter):
    def __init__(self, p_expression):
        super().__init__(p_expression, _ORDINAL_TAG_MATCH)

    def match(self, p_todo):
        """
        Performs a match on a key:value tag in the todo.

        First it tries to convert the value and the user-entered expression to
        a date and makes a comparison if it succeeds, based on the given
        operator (default ==).
        Upon failure, it falls back to converting value and user-entered
        expression to an integer and makes a numerical comparison based on the
        given operator (default ==)
        As a last resort, it falls back to using a Grep filter to see if the
        user given expression is contained in the todo text.
        """
        def resort_to_grep_filter():
            grep = GrepFilter(self.expression)
            return grep.match(p_todo)

        if not self.key or not p_todo.has_tag(self.key):
            return False

        if len(p_todo.tag_values(self.key)) > 1:
            # we cannot apply an ordening on a tag that appears more than once
            # in this todo item, therefore use a simple value match
            return resort_to_grep_filter()

        try:
            operand1 = date_string_to_date(p_todo.tag_value(self.key))
            operand2 = relative_date_to_date(self.value)

            if not operand2:
                operand2 = date_string_to_date(self.value)

        except ValueError:
            operand1 = p_todo.tag_value(self.key)
            operand2 = self.value

            try:
                operand1 = int(operand1)
                operand2 = int(operand2)
            except ValueError:
                return resort_to_grep_filter()

        return self.compare_operands(operand1, operand2)


class _DateAttributeFilter(OrdinalFilter):
    def __init__(self, p_expression, p_match, p_getter):
        super().__init__(p_expression, p_match)
        self.getter = p_getter

    def match(self, p_todo):
        operand1 = self.getter(p_todo)
        operand2 = relative_date_to_date(self.value)

        if not operand2:
            operand2 = date_string_to_date(self.value)

        if operand1 and operand2:
            return self.compare_operands(operand1, operand2)
        else:
            return False


_CREATED_MATCH = r'creat(ion|ed?):' + _OPERATOR_MATCH + _VALUE_MATCH


class CreationFilter(_DateAttributeFilter):
    def __init__(self, p_expression):
        super().__init__(
            p_expression,
            _CREATED_MATCH,
            lambda t: t.creation_date()  # pragma: no branch
        )


_COMPLETED_MATCH = r'complet(ed?|ion):' + _OPERATOR_MATCH + _VALUE_MATCH


class CompletionFilter(_DateAttributeFilter):
    def __init__(self, p_expression):
        super().__init__(
            p_expression,
            _COMPLETED_MATCH,
            lambda t: t.completion_date()  # pragma: no branch
        )


_PRIORITY_MATCH = r"\(" + _OPERATOR_MATCH + r"(?P<value>[A-Z]{1})\)"


class PriorityFilter(OrdinalFilter):
    def __init__(self, p_expression):
        super().__init__(p_expression, _PRIORITY_MATCH)

    def match(self, p_todo):
        """
        Performs a match on a priority in the todo.

        It gets priority from p_todo and compares it with user-entered
        expression based on the given operator (default ==). It does that however
        in reversed order to obtain more intuitive result. Example: (>B) will
        match todos with priority (A).
        Items without priority are designated with corresponding operand set to
        'ZZ', because python doesn't allow NoneType() and str() comparisons.
        """
        operand1 = self.value
        operand2 = p_todo.priority() or 'ZZ'

        return self.compare_operands(operand1, operand2)

MATCHES = [
    (_CREATED_MATCH, CreationFilter),
    (_COMPLETED_MATCH, CompletionFilter),
    (_ORDINAL_TAG_MATCH, OrdinalTagFilter),
    (_PRIORITY_MATCH, PriorityFilter),
]

def get_filter_list(p_expression):
    """
    Returns a list of GrepFilters, OrdinalTagFilters or NegationFilters based
    on the given filter expression.

    The filter expression is a list of strings.
    """
    result = []
    for arg in p_expression:
        # when a word starts with -, it should be negated
        is_negated = len(arg) > 1 and arg[0] == '-'
        arg = arg[1:] if is_negated else arg

        argfilter = None
        for match, _filter in MATCHES:
            if re.match(match, arg):
                argfilter = _filter(arg)
                break

        if not argfilter:
            argfilter = GrepFilter(arg)

        if is_negated:
            argfilter = NegationFilter(argfilter)

        result.append(argfilter)

    return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Provides a pretty printer filter that colorizes todo items. """

import re

from topydo.lib.Color import AbstractColor
from topydo.lib.Config import config
from topydo.lib.PrettyPrinterFilter import PrettyPrinterFilter
from topydo.lib.TopydoString import TopydoString


class PrettyPrinterColorFilter(PrettyPrinterFilter):
    """
    Adds colors to the todo string by inserting ANSI codes.

    Should be passed as a filter in the filter list of pretty_print()
    """

    def filter(self, p_todo_str, p_todo):
        """ Applies the colors. """
        if config().colors():
            p_todo_str = TopydoString(p_todo_str, p_todo)

            priority_color = config().priority_color(p_todo.priority())

            colors = [
                (r'\B@(\S*\w)', AbstractColor.CONTEXT),
                (r'\B\+(\S*\w)', AbstractColor.PROJECT),
                (r'\b\S+:[^/\s]\S*\b', AbstractColor.META),
                (r'(^|\s)(\w+:){1}(//\S+)', AbstractColor.LINK),
            ]

            # color by priority
            p_todo_str.set_color(0, priority_color)

            for pattern, color in colors:
                for match in re.finditer(pattern, p_todo_str.data):
                    p_todo_str.set_color(match.start(), color)
                    p_todo_str.set_color(match.end(), priority_color)

            p_todo_str.append('', AbstractColor.NEUTRAL)

        return p_todo_str



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Provides a base class for pretty printer filters. """


class PrettyPrinterFilter(object):
    """
    Base class for a pretty printer filter.

    Subclasses must re-implement the filter method.
    """

    def filter(self, p_todo_str, _):
        """
        Applies a filter to p_todo_str and returns a modified version of it.
        """
        raise NotImplementedError



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" A view is a list of todos, sorted, grouped and filtered. """


class View(object):
    """
    A view is instantiated by a todo list, usually obtained from a todo.txt
    file. Also a sorter and a list of filters should be given that is applied
    to the list.
    """

    def __init__(self, p_sorter, p_filters, p_todolist):
        self.todolist = p_todolist
        self._sorter = p_sorter
        self._filters = p_filters

    def _apply_filters(self, p_todos):
        """ Applies the filters to the list of todo items. """
        result = p_todos

        for _filter in sorted(self._filters, key=lambda f: f.order):
            result = _filter.filter(result)

        return result

    @property
    def todos(self):
        """ Returns a sorted and filtered list of todos in this view. """
        result = self._sorter.sort(self.todolist.todos())
        return self._apply_filters(result)

    @property
    def groups(self):
        result = self._apply_filters(self.todolist.todos())
        return self._sorter.group(result)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Various utility functions.
"""

import re
from collections import namedtuple
from datetime import date

import arrow


def date_string_to_date(p_date):
    """
    Given a date in YYYY-MM-DD, returns a Python date object. Throws a
    ValueError if the date is invalid.
    """
    result = None

    if p_date:
        parsed_date = re.match(r'(\d{4})-(\d{2})-(\d{2})', p_date)
        if parsed_date:
            result = date(
                int(parsed_date.group(1)),  # year
                int(parsed_date.group(2)),  # month
                int(parsed_date.group(3))   # day
            )
        else:
            raise ValueError

    return result


def is_valid_priority(p_priority):
    return p_priority is not None and re.match(r'^[A-Z]$', p_priority) is not None


def escape_ansi(p_string):
    return escape_ansi.pattern.sub('', p_string)

escape_ansi.pattern = re.compile(r'\x1b[^m]*m')


def get_terminal_size(p_getter=None):
    """
    Try to determine terminal size at run time. If that is not possible,
    returns the default size of 80x24.

    By default, the size is determined with provided get_terminal_size by
    shutil. Sometimes an UI may want to specify the desired width, then it can
    provide a getter that returns a named tuple (columns, lines) with the size.
    """

    try:
        return get_terminal_size.getter()
    except AttributeError:
        if p_getter:
            get_terminal_size.getter = p_getter
        else:
            def inner():
                try:
                    # shutil.get_terminal_size was added to the standard
                    # library in Python 3.3
                    from shutil import \
                        get_terminal_size as \
                        _get_terminal_size  # pylint: disable=no-name-in-module

                    size = _get_terminal_size()
                except ValueError:
                    # This can result from the 'underlying buffer being detached', which
                    # occurs during running the unittest on Windows (but not on Linux?)
                    terminal_size = namedtuple('Terminal_Size', 'columns lines')
                    size = terminal_size(80, 24)

                return size

            get_terminal_size.getter = inner

        return get_terminal_size.getter()


def translate_key_to_config(p_key):
    """
    Translates urwid key event to form understandable by topydo config parser.
    """
    if len(p_key) > 1:
        key = p_key.capitalize()
        if key.startswith('Ctrl') or key.startswith('Meta'):
            key = key[0] + '-' + key[5:]
        key = '<' + key + '>'
    else:
        key = p_key

    return key

def humanize_date(p_datetime):
    """ Returns a relative date string from a datetime object. """
    now = arrow.now()
    _date = now.replace(day=p_datetime.day, month=p_datetime.month, year=p_datetime.year)
    return _date.humanize(now).replace('just now', 'today')



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import Command, InvalidCommandArgument
from topydo.lib.Config import config
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.TodoListBase import InvalidTodoException


class TagCommand(Command):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.force = False
        self.force_add = False
        self.relative_date = False
        self.todo = None
        self.tag = None
        self.value = None
        self.values = []
        self.current_values = []

    def _process_flags(self):
        flags, args = self.getopt("afr")
        for flag, _ in flags:
            if flag == "-a":
                self.force_add = True
            elif flag == "-f":
                self.force = True
            elif flag == "-r":
                self.relative_date = True

        self.args = args

    def _process_args(self):
        self._process_flags()

        try:
            self.todo = self.todolist.todo(self.argument(0))
            self.tag = self.argument(1)
            self.current_values = self.todo.tag_values(self.tag)
        except InvalidTodoException:
            self.error("Invalid todo number.")
        except InvalidCommandArgument:
            self.error(self.usage())

        try:
            self.value = self.argument(2)
        except InvalidCommandArgument:
            self.value = ""

    def _print(self):
        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))
        self.out(self.printer.print_todo(self.todo))

    def _choose(self):
        """
        Returns the chosen number of the tag value to process (or "all").
        """
        answer = "all"

        if not self.force:
            for i, value in enumerate(self.current_values):
                self.out("{:>2d}. {}".format(i + 1, value))

            answer = self.prompt(
                'Which value to remove? Enter number or "all": ')

        if answer != "all":
            try:
                answer = int(answer) - 1

                if answer < 0 or answer >= len(self.current_values):
                    answer = None
            except ValueError:
                answer = None

        return answer

    def _convert_relative_dates(self):
        is_start_tag = self.tag == config().tag_start()
        is_due_tag = self.tag == config().tag_due()

        if self.relative_date or is_start_tag or is_due_tag:
            real_date = relative_date_to_date(self.value)

            if real_date:
                self.value = real_date.isoformat()

    def _set_helper(self, p_old_value=""):
        self._convert_relative_dates()

        old_src = self.todo.source()
        self.todo.set_tag(self.tag, self.value, self.force_add, p_old_value)

        if old_src != self.todo.source():
            self.todolist.dirty = True

    def _set(self):
        if len(self.current_values) > 1 and not self.force_add:
            answer = self._choose()

            if answer == "all":
                for value in self.current_values:
                    self._set_helper(value)
            elif answer is not None and self.value != self.current_values[answer]:
                self._set_helper(self.current_values[answer])

        else:
            self._set_helper()

        self._print()

    def execute(self):
        if not super().execute():
            return False

        self._process_args()

        if self.todo and self.tag:
            self._set()

    def usage(self):
        return """Synopsis: tag [-a] [-f] [-r] <NUMBER> <TAG> [<VALUE>]"""

    def help(self):
        return """\
Sets the given TAG on the given todo NUMBER with the given VALUE. If the VALUE
is omitted, the TAG is removed from the todo item.

-a : Do not change the current value of the TAG if it exists, but add a new
     VALUE for the given TAG.
-f : Force setting/removing all values of the TAG. Prevents interaction with
     the user.
-r : Interpret the given value as a relative date and convert it to an absolute
     date.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2016 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module deals with todo.txt files while putting a watch on them for file
changes.
"""

import os.path

from watchdog.events import (FileCreatedEvent, FileModifiedEvent,
                             FileSystemEventHandler)
from watchdog.observers import Observer

from topydo.lib.TodoFile import TodoFile


class TodoFileWatched(TodoFile):
    """
    This class represents a todo.txt file, which can be read from or written
    to.
    """

    def __init__(self, p_path, p_on_update):
        super().__init__(p_path)
        self.self_write = False

        class EventHandler(FileSystemEventHandler):
            """
            Event handler to catch modifications (or creations) of the
            current todo.txt file.
            """
            def __init__(self, p_file):
                super().__init__()
                self.file = p_file

            def _handle(self, p_event):
                right_type = isinstance(p_event, FileModifiedEvent) or isinstance(p_event, FileCreatedEvent)
                should_trigger = right_type and p_event.src_path == self.file.path

                if self.file.self_write and should_trigger:
                    # the file was written by topydo, unmark that so we can
                    # record external writes again.
                    self.file.self_write = False
                elif should_trigger:
                    p_on_update()

            def on_created(self, p_event):
                """
                Because vim deletes and creates a file on buffer save, also
                catch a creation event.
                """
                self._handle(p_event)

            def on_modified(self, p_event):
                self._handle(p_event)

        observer = Observer()
        observer.schedule(EventHandler(self), os.path.dirname(self.path))
        observer.start()

    def write(self, p_todos):
        # make sure not to reread the todo file because this instance is
        # actually writing it
        self.self_write = True
        super().write(p_todos)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Provides a pretty printer filter that inserts todo numbers. """

from topydo.lib.PrettyPrinterFilter import PrettyPrinterFilter


class PrettyPrinterNumbers(PrettyPrinterFilter):
    """ Prepends the todo's number, retrieved from the todolist. """

    def __init__(self, p_todolist):
        super().__init__()
        self.todolist = p_todolist

    def filter(self, p_todo_str, p_todo):
        """ Prepends the number to the todo string. """
        return "|{:>3}| {}".format(self.todolist.number(p_todo), p_todo_str)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from topydo.lib.Command import Command


class ExitCommand(Command):
    """
    A command that exits topydo. Used for the 'exit' and 'quit' subcommands on
    the prompt CLI.
    """

    def __init__(self, p_args, p_todolist, p_output, p_error, p_input):
        super().__init__(p_args, p_todolist, p_output, p_error,
                                          p_input)

    def execute(self):
        if not super().execute():
            return False

        sys.exit(0)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
A list of todo items.
"""

import types

from topydo.lib.Config import config
from topydo.lib.TodoListBase import TodoListBase


def _needs_dependencies(p_function):
    """
    A decorator that triggers the population of the dependency tree in a
    TodoList (and other administration). The decorator should be applied to
    methods of TodoList that require dependency information.
    """
    def build_dependency_information(p_todolist):
        for todo in p_todolist._todos:
            p_todolist._register_todo(todo)

    def inner(self, *args, **kwargs):
        if not self._initialized:
            self._initialized = True

            from topydo.lib.Graph import DirectedGraph
            self._depgraph = DirectedGraph()

            build_dependency_information(self)

        return p_function(self, *args, **kwargs)

    return inner

class TodoList(TodoListBase):
    """
    Provides operations for a todo list, such as adding items, removing them,
    etc.

    The list is usually a complete list found in the program's input (e.g. a
    todo.txt file), not an arbitrary set of todo items.
    """

    def __init__(self, p_todostrings):
        """
        Should be given a list of strings, each element a single todo string.
        The string will be parsed.
        """
        self._initialized = False  # whether dependency information was
                                   # initialized

        # initialize these first because the constructor calls add_list
        self._tododict = {}  # hash(todo) to todo lookup
        self._parentdict = {}  # dependency id => parent todo
        self._depgraph = None

        super().__init__(p_todostrings)

    @_needs_dependencies
    def todo_by_dep_id(self, p_dep_id):
        """
        Returns the todo that has the id tag set to the value p_dep_id.
        There is only one such task, the behavior is undefined when a todo item
        has more than one id tag.
        """
        try:
            return self._parentdict[p_dep_id]
        except KeyError:
            return None

    def _maintain_dep_graph(self, p_todo):
        """
        Makes sure that the dependency graph is consistent according to the
        given todo.
        """
        dep_id = p_todo.tag_value('id')
        # maintain dependency graph
        if dep_id:
            self._parentdict[dep_id] = p_todo
            self._depgraph.add_node(hash(p_todo))

            # connect all tasks we have in memory so far that refer to this
            # task
            for dep in \
                    [dep for dep in self._todos if dep.has_tag('p', dep_id)]:

                self._add_edge(p_todo, dep, dep_id)

        for dep_id in p_todo.tag_values('p'):
            try:
                parent = self._parentdict[dep_id]
                self._add_edge(parent, p_todo, dep_id)
            except KeyError:
                pass

    def _register_todo(self, p_todo):
        self._maintain_dep_graph(p_todo)
        self._tododict[hash(p_todo)] = p_todo

    def add_todos(self, p_todos):
        super().add_todos(p_todos)

        for todo in self._todos:
            todo.parents = types.MethodType(self.parents, todo)

            # only do administration when the dependency info is initialized,
            # otherwise we postpone it until it's really needed (through the
            # _needs_dependencies decorator)
            if self._initialized:
                self._register_todo(todo)

    def delete(self, p_todo, p_leave_tags=False):
        """ Deletes a todo item from the list. """
        try:
            number = self._todos.index(p_todo)

            if p_todo.has_tag('id'):
                for child in self.children(p_todo):
                    self.remove_dependency(p_todo, child, p_leave_tags)

            if p_todo.has_tag('p'):
                for parent in self.parents(p_todo):
                    self.remove_dependency(parent, p_todo, p_leave_tags)

            del self._todos[number]
            self._update_todo_ids()

            self.dirty = True
        except ValueError:
            # todo item couldn't be found, ignore
            pass

    def _add_edge(self, p_from_todo, p_to_todo, p_dep_id):
        self._parentdict[p_dep_id] = p_from_todo
        self._depgraph.add_edge(hash(p_from_todo), hash(p_to_todo), p_dep_id)

    @_needs_dependencies
    def add_dependency(self, p_from_todo, p_to_todo):
        """ Adds a dependency from task 1 to task 2. """
        def find_next_id():
            """
            Find a new unused ID.
            Unused means that no task has it as an 'id' value or as a 'p'
            value.
            """
            def id_exists(p_id):
                """
                Returns True if there exists a todo with the given parent ID.
                """
                for todo in self._todos:
                    number = str(p_id)
                    if todo.has_tag('id', number) or todo.has_tag('p', number):
                        return True

                return False

            new_id = 1
            while id_exists(new_id):
                new_id += 1

            return str(new_id)

        def append_projects_to_subtodo():
            """
            Appends projects in the parent todo item that are not present in
            the sub todo item.
            """
            if config().append_parent_projects():
                for project in p_from_todo.projects() - p_to_todo.projects():
                    self.append(p_to_todo, "+{}".format(project))

        def append_contexts_to_subtodo():
            """
            Appends contexts in the parent todo item that are not present in
            the sub todo item.
            """
            if config().append_parent_contexts():
                for context in p_from_todo.contexts() - p_to_todo.contexts():
                    self.append(p_to_todo, "@{}".format(context))

        if p_from_todo != p_to_todo and not self._depgraph.has_edge(
                hash(p_from_todo), hash(p_to_todo)):

            dep_id = None
            if p_from_todo.has_tag('id'):
                dep_id = p_from_todo.tag_value('id')
            else:
                dep_id = find_next_id()
                p_from_todo.set_tag('id', dep_id)

            p_to_todo.add_tag('p', dep_id)
            self._add_edge(p_from_todo, p_to_todo, dep_id)
            append_projects_to_subtodo()
            append_contexts_to_subtodo()
            self.dirty = True

    @_needs_dependencies
    def remove_dependency(self, p_from_todo, p_to_todo, p_leave_tags=False):
        """ Removes a dependency between two todos. """
        dep_id = p_from_todo.tag_value('id')

        if dep_id:
            self._depgraph.remove_edge(hash(p_from_todo), hash(p_to_todo))
            self.dirty = True

        # clean dangling dependency tags
        if dep_id and not p_leave_tags:
            p_to_todo.remove_tag('p', dep_id)

            if not self.children(p_from_todo, True):
                p_from_todo.remove_tag('id')
                del self._parentdict[dep_id]

    @_needs_dependencies
    def parents(self, p_todo, p_only_direct=False):
        """
        Returns a list of parent todos that (in)directly depend on the
        given todo.
        """
        parents = self._depgraph.incoming_neighbors(
            hash(p_todo), not p_only_direct)
        return [self._tododict[parent] for parent in parents]

    @_needs_dependencies
    def children(self, p_todo, p_only_direct=False):
        """
        Returns a list of child todos that the given todo (in)directly depends
        on.
        """
        children = \
            self._depgraph.outgoing_neighbors(hash(p_todo), not p_only_direct)
        return [self._tododict[child] for child in children]

    @_needs_dependencies
    def clean_dependencies(self):
        """
        Cleans the dependency graph.

        This is achieved by performing a transitive reduction on the dependency
        graph and removing unused dependency ids from the graph (in that
        order).
        """
        def remove_tag(p_todo, p_tag, p_value):
            """
            Removes a tag from a todo item.
            """
            p_todo.remove_tag(p_tag, p_value)
            self.dirty = True

        def clean_parent_relations():
            """
            Remove id: tags for todos without child todo items.
            """

            for todo in [todo for todo in self._todos if todo.has_tag('id')]:
                value = todo.tag_value('id')
                if not self._depgraph.has_edge_id(value):
                    remove_tag(todo, 'id', value)
                    del self._parentdict[value]

        def clean_orphan_relations():
            """
            Remove p: tags for todos referring to a parent that is not in the
            dependency graph anymore.
            """

            for todo in [todo for todo in self._todos if todo.has_tag('p')]:
                for value in todo.tag_values('p'):
                    parent = self.todo_by_dep_id(value)

                    if not self._depgraph.has_edge(hash(parent), hash(todo)):
                        remove_tag(todo, 'p', value)

        self._depgraph.transitively_reduce()
        clean_parent_relations()
        clean_orphan_relations()



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

import urwid

from topydo.lib.Config import config
from topydo.lib.ListFormat import ListFormatParser
from topydo.lib.ProgressColor import progress_color
from topydo.ui.columns.Utils import PaletteItem, to_urwid_color

# pass a None todo list, since we won't use %i or %I here
PRIO_FORMATTER = ListFormatParser(None, "%{(}p{)}")
TEXT_FORMATTER = ListFormatParser(None, "%s %k\n%h")

PRJ_CON_PATTERN = r'\B(?:\+|@)(?:\S*\w)'
TAG_PATTERN = r'\b\S+:[^/\s]\S*\b'
URL_PATTERN = r'(?:^|\s)(?:\w+:){1}(?://\S+)'


def _markup(p_todo, p_focus):
    """
    Returns an attribute spec for the colors that correspond to the given todo
    item.
    """
    pri = p_todo.priority()
    pri = 'pri_' + pri if pri else PaletteItem.DEFAULT

    if not p_focus:
        attr_dict = {None: pri}
    else:
        # use '_focus' palette entries instead of standard ones
        attr_dict = {None: pri + '_focus'}
        attr_dict[PaletteItem.PROJECT] = PaletteItem.PROJECT_FOCUS
        attr_dict[PaletteItem.CONTEXT] = PaletteItem.CONTEXT_FOCUS
        attr_dict[PaletteItem.METADATA] = PaletteItem.METADATA_FOCUS
        attr_dict[PaletteItem.LINK] = PaletteItem.LINK_FOCUS

    return attr_dict


class TodoWidget(urwid.WidgetWrap):
    def __init__(self, p_todo, p_id_width=4):
        # clients use this to associate this widget with the given todo item
        self.todo = p_todo

        todo_text = TEXT_FORMATTER.parse(p_todo)

        if p_todo.is_completed():
            priority_text = ' x '
        else:
            priority_text = PRIO_FORMATTER.parse(p_todo)

        # split todo_text at each occurrence of tag/project/context/url
        txt_pattern = r'|'.join([PRJ_CON_PATTERN, TAG_PATTERN, URL_PATTERN])
        txt_pattern = r'(' + txt_pattern + r')'
        txt_splitted = re.split(txt_pattern, todo_text)
        txt_markup = []

        # Examine each substring and apply relevant palette entry if needed
        for substring in txt_splitted:
            # re.split can generate empty strings when capturing group is used
            if not substring:
                continue
            if re.match(TAG_PATTERN, substring):
                txt_markup.append((PaletteItem.METADATA, substring))
            elif re.match(URL_PATTERN, substring):
                txt_markup.append((PaletteItem.LINK, substring))
            elif re.match(PRJ_CON_PATTERN, substring):
                if substring.startswith('+'):
                    txt_markup.append((PaletteItem.PROJECT, substring))
                else:
                    txt_markup.append((PaletteItem.CONTEXT, substring))
            else:
                txt_markup.append(substring)

        self.id_widget = urwid.Text('', align='right')
        priority_widget = urwid.Text(priority_text)
        self.text_widget = urwid.Text(txt_markup)

        self.progress_bar = urwid.AttrMap(
                urwid.SolidFill(' '),
                {},
        )
        self.update_progress()

        self.columns = urwid.Columns(
            [
                (1, self.progress_bar),
                (p_id_width, self.id_widget),
                (3, priority_widget),
                ('weight', 1, self.text_widget),
            ],
            dividechars=1,
            box_columns=[0] # the progress bar adapts its height to the rest
        )

        self.widget = urwid.AttrMap(
            self.columns,
            _markup(p_todo, p_focus=False),
            _markup(p_todo, p_focus=True)
        )

        super().__init__(self.widget)

    # pylint: disable=no-self-use
    def keypress(self, _, p_key):
        """
        Override keypress to prevent the wrapped Columns widget to
        receive any key.
        """
        return p_key

    # pylint: disable=no-self-use
    def selectable(self):
        # make sure that ListBox will highlight this widget
        return True

    @property
    def number(self):
        pass

    @number.setter
    def number(self, p_number):
        self.id_widget.set_text(str(p_number))

    def update_progress(self):
        color = to_urwid_color(progress_color(self.todo)) if config().colors() else PaletteItem.DEFAULT

        self.progress_bar.set_attr_map(
            {None: urwid.AttrSpec(PaletteItem.DEFAULT, color, 256)}
        )

    def mark(self):
        attr_map = {
            None:                 PaletteItem.MARKED,
            PaletteItem.LINK:     PaletteItem.MARKED,
            PaletteItem.CONTEXT:  PaletteItem.MARKED,
            PaletteItem.PROJECT:  PaletteItem.MARKED,
            PaletteItem.METADATA: PaletteItem.MARKED,
        }
        self.widget.set_attr_map(attr_map)

    def unmark(self):
        self.widget.set_attr_map(_markup(self.todo, False))

    cache = {}

    @classmethod
    def create(p_class, p_todo, p_id_width=4):
        """
        Creates a TodoWidget instance for the given todo. Widgets are
        cached, the same object is returned for the same todo item.
        """

        def parent_progress_may_have_changed(p_todo):
            """
            Returns True when a todo's progress should be updated because it is
            dependent on the parent's progress.
            """
            return p_todo.has_tag('p') and not p_todo.has_tag('due')

        source = p_todo.source()

        if source in p_class.cache:
            widget = p_class.cache[source]

            if p_todo is not widget.todo:
                # same source text but different todo instance (could happen
                # after an edit where a new Todo instance is created with the
                # same text as before)
                # simply fix the reference in the stored widget.
                widget.todo = p_todo

            if parent_progress_may_have_changed(p_todo):
                widget.update_progress()
        else:
            widget = p_class(p_todo, p_id_width)
            p_class.cache[source] = widget

        return widget

    @classmethod
    def wipe_cache(p_class):
        """ Wipes the cache """
        p_class.cache = {}



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides a printer that transforms a list of Todo items to a JSON structure
such that other applications can process it.
"""

import json

from topydo.lib.printers.PrettyPrinter import Printer


def _convert_todo(p_todo):
    """ Converts a Todo instance to a dictionary. """
    creation_date = p_todo.creation_date()
    completion_date = p_todo.completion_date()

    result = {
        'source': p_todo.source(),
        'text': p_todo.text(),
        'priority': p_todo.priority(),
        'completed': p_todo.is_completed(),
        'tags': p_todo.tags(),
        'projects': list(p_todo.projects()),
        'contexts': list(p_todo.contexts()),
        'creation_date':
            creation_date.isoformat() if creation_date else None,
        'completion_date':
            completion_date.isoformat() if completion_date else None
    }

    return result


class JsonPrinter(Printer):
    """
    A printer that converts a list of Todo items to a string in JSON format.
    """

    def __init__(self):
        super().__init__()

    def print_todo(self, p_todo):
        return json.dumps(_convert_todo(p_todo), ensure_ascii=False,
                          sort_keys=True)

    def print_list(self, p_todos):
        result = []

        for todo in p_todos:
            result.append(_convert_todo(todo))

        return json.dumps(result, ensure_ascii=False, sort_keys=True)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module deals with todo.txt files.
"""

import codecs
import os.path


class TodoFile(object):
    """
    This class represents a todo.txt file, which can be read from or written
    to.
    """

    def __init__(self, p_path):
        self.path = os.path.abspath(p_path)

    def read(self):
        """ Reads the todo.txt file and returns a list of todo items. """
        todos = []
        try:
            todofile = codecs.open(self.path, 'r', encoding="utf-8")
            todos = todofile.readlines()
            todofile.close()
        except IOError:
            pass

        return todos

    def write(self, p_todos):
        """
        Writes all the todo items to the todo.txt file.

        p_todos can be a list of todo items, or a string that is just written
        to the file.
        """

        todofile = codecs.open(self.path, 'w', encoding="utf-8")

        if p_todos is list:
            for todo in p_todos:
                todofile.write(str(todo))
        else:
            todofile.write(p_todos)

        todofile.write("\n")

        todofile.close()



""" Version of Topydo. """

VERSION = '0.16'

LICENSE = """Copyright (C) 2014-2017 Bram Schoenmakers
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains a base class for a CLI implementation of topydo and functions for the
I/O on the command-line.
"""

import getopt
import sys

from topydo.lib.Color import AbstractColor, Color
from topydo.lib.TopydoString import TopydoString

MAIN_OPTS = "ac:C:d:ht:v"
MAIN_LONG_OPTS = ('version')
READ_ONLY_COMMANDS = ('list', 'listcontext', 'listproject')

GENERIC_HELP="""Available commands:

* add
* append (app)
* del (rm)
* dep
* depri
* do
* edit
* ls
* listcon (lscon)
* listprojects (lsprj)
* modify (mod)
* postpone
* pri
* revert
* sort
* tag

Run `topydo help <subcommand>` for command-specific help.\
"""

def usage():
    """ Prints the command-line usage of topydo. """

    print("""\
Synopsis: topydo [-a] [-c <config>] [-C <colormode>] [-d <archive>] [-t <todo.txt>] subcommand [help|args]
          topydo -h
          topydo -v

-a : Do not archive todo items on completion.
-c : Specify an alternative configuration file.
-C : Specify color mode (0 = disable, 1 = enable 16 colors,
     16 = enable 16 colors, 256 = enable 256 colors, auto (default))
-d : Specify an alternative archive file (done.txt)
-h : This help text
-t : Specify an alternative todo file
-v : Print version and exit

""" + GENERIC_HELP)

def write(p_file, p_string):
    """
    Write p_string to file p_file, trailed by a newline character.

    ANSI codes are removed when the file is not a TTY (and colors are
    automatically determined).
    """
    if not config().colors(p_file.isatty()):
        p_string = escape_ansi(p_string)

    if p_string:
        p_file.write(p_string + "\n")


def lookup_color(p_color):
    """
    Converts an AbstractColor to a normal Color. Returns the Color itself
    when a normal color is passed.
    """
    if not lookup_color.colors:
        lookup_color.colors[AbstractColor.NEUTRAL] = Color('NEUTRAL')
        lookup_color.colors[AbstractColor.PROJECT] = config().project_color()
        lookup_color.colors[AbstractColor.CONTEXT] = config().context_color()
        lookup_color.colors[AbstractColor.META] = config().metadata_color()
        lookup_color.colors[AbstractColor.LINK] = config().link_color()

    try:
        return lookup_color.colors[p_color]
    except KeyError:
        return p_color

lookup_color.colors = {}

def insert_ansi(p_string):
    """ Returns a string with color information at the right positions.  """
    result = p_string.data

    for pos, color in sorted(p_string.colors.items(), reverse=True):
        color = lookup_color(color)

        result = result[:pos] + color.as_ansi() + result[pos:]

    return result

def output(p_string):
    if isinstance(p_string, list):
        p_string = "\n".join([insert_ansi(s) for s in p_string])
    elif isinstance(p_string, TopydoString):
        # convert color codes to ANSI
        p_string = insert_ansi(p_string)

    write(sys.stdout, p_string)

def error(p_string):
    """ Writes an error on the standard error. """
    write(sys.stderr, p_string)


def version():
    """ Print the current version and exit. """
    from topydo.lib.Version import VERSION, LICENSE
    print("topydo {}\n".format(VERSION))
    print(LICENSE)
    sys.exit(0)

from topydo.lib.Config import config, ConfigError

# First thing is to poke the configuration and check whether it's sane
# The modules below may already read in configuration upon import, so
# make sure to bail out if the configuration is invalid.
try:
    config()
except ConfigError as config_error:
    error(str(config_error))
    sys.exit(1)

from topydo.lib import TodoFile
from topydo.lib import TodoList
from topydo.lib import TodoListBase
from topydo.lib.Utils import escape_ansi


def _retrieve_archive():
    """
    Returns a tuple with archive content: the first element is a TodoListBase
    and the second element is a TodoFile.
    """
    archive_file = TodoFile.TodoFile(config().archive())
    archive = TodoListBase.TodoListBase(archive_file.read())

    return (archive, archive_file)


class CLIApplicationBase(object):
    """
    Base class for a Command Line Interfaces (CLI) for topydo. Examples are the
    original CLI and the Prompt interface.

    Handles input/output of the various subcommands.
    """

    def __init__(self):
        self.todolist = TodoList.TodoList([])
        self.todofile = None
        self.do_archive = True
        self._post_archive_action = None
        self.backup = None

    @staticmethod
    def _usage():
        usage()
        sys.exit(0)

    def _process_flags(self):
        args = sys.argv[1:]

        try:
            opts, args = getopt.getopt(args, MAIN_OPTS, MAIN_LONG_OPTS)
        except getopt.GetoptError as e:
            error(str(e))
            sys.exit(1)

        alt_config_path = None
        overrides = {}

        for opt, value in opts:
            if opt == "-a":
                self.do_archive = False
            elif opt == "-c":
                alt_config_path = value
            elif opt == "-C":
                overrides[('topydo', 'force_colors')] = '1'
                overrides[('topydo', 'colors')] = value
            elif opt == "-t":
                overrides[('topydo', 'filename')] = value
            elif opt == "-d":
                overrides[('topydo', 'archive_filename')] = value
            elif opt in ("-v", "--version"):
                version()
            else:
                CLIApplicationBase._usage()

        if alt_config_path:
            config(alt_config_path, overrides)
        elif len(overrides):
            config(p_overrides=overrides)

        return args

    def _archive(self):
        """
        Performs an archive action on the todolist.

        This means that all completed tasks are moved to the archive file
        (defaults to done.txt).
        """
        archive, archive_file = _retrieve_archive()

        if self.backup:
            self.backup.add_archive(archive)

        if archive:
            from topydo.commands.ArchiveCommand import ArchiveCommand
            command = ArchiveCommand(self.todolist, archive)
            command.execute()

            if archive.dirty:
                archive_file.write(archive.print_todos())

    @staticmethod
    def is_read_only(p_command):
        """ Returns True when the given command class is read-only. """
        read_only_commands = tuple(cmd for cmd
                                   in ('revert', ) + READ_ONLY_COMMANDS)
        return p_command.name() in read_only_commands

    def _backup(self, p_command, p_args=None, p_label=None):
        if config().backup_count() > 0 and p_command and not CLIApplicationBase.is_read_only(p_command):
            p_args = p_args if p_args else []
            call = [p_command.name()] + p_args

            from topydo.lib.ChangeSet import ChangeSet
            label = p_label if p_label else call
            self.backup = ChangeSet(self.todolist, p_label=label)

    def _execute(self, p_command, p_args):
        """
        Execute a subcommand with arguments. p_command is a class (not an
        object).
        """
        self._backup(p_command, p_args)

        command = p_command(
            p_args,
            self.todolist,
            output,
            error,
            input)

        if command.execute() != False:
            self._post_archive_action = command.execute_post_archive_actions
            return True

        return False

    def _post_execute(self):
        """
        Should be called when executing the user requested command has been
        completed. It will do some maintenance and write out the final result
        to the todo.txt file.
        """

        if self.todolist.dirty:
            # do not archive when the value of the filename is an empty string
            # (i.e. explicitly left empty in the configuration
            if self.do_archive and config().archive():
                self._archive()
            elif config().archive() and self.backup:
                archive = _retrieve_archive()[0]
                self.backup.add_archive(archive)

            self._post_archive_action()

            if config().keep_sorted():
                from topydo.commands.SortCommand import SortCommand
                self._execute(SortCommand, [])

            if self.backup:
                self.backup.save(self.todolist)

            self.todofile.write(self.todolist.print_todos())
            self.todolist.dirty = False

        self.backup = None

    def run(self):
        raise NotImplementedError



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2017 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from topydo.ui.columns.Utils import PaletteItem


class CompletionBoxWidget(urwid.ListBox):
    def __init__(self):
        self.items = urwid.SimpleFocusListWalker([])
        self.min_width = 0

        super().__init__(self.items)

    def __len__(self):
        return len(self.items)

    @property
    def height(self):
        """ Returns height of the widget, with maximum set to 4 lines. """
        return min(len(self), 4)

    @property
    def margin(self):
        """
        Returns margin for rendering the widget always glued to the cursor.
        """
        return len(self.items[0].original_widget.text)

    def add_completions(self, p_completions):
        """
        Creates proper urwid.Text widgets for all completion candidates from
        p_completions list, and populates them into the items attribute.
        """
        palette = PaletteItem.MARKED
        for completion in p_completions:
            width = len(completion)
            if width > self.min_width:
                self.min_width = width
            w = urwid.Text(completion)
            self.items.append(urwid.AttrMap(w, None, focus_map=palette))
            self.items.set_focus(0)

    def clear(self):
        self.items.clear()
        self.min_width = 0

    def set_focus(self, p_position, p_coming_from=None):
        self.focus.set_attr_map({PaletteItem.MARKED: None})
        super().set_focus(p_position, p_coming_from)
        self.focus.set_attr_map({None: PaletteItem.MARKED})



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.MultiCommand import MultiCommand
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers


class DepriCommand(MultiCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

    def _execute_multi_specific(self):
        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))

        for todo in self.todos:
            if todo.priority() != None:
                self.todolist.set_priority(todo, None)
                self.out("Priority removed.")
                self.out(self.printer.print_todo(todo))

    def usage(self):
        return """\
Synopsis: depri <NUMBER 1> [<NUMBER 2> ...]
          depri [-x] -e <EXPRESSION>\
"""

    def help(self):
        return """Removes the priority of the given todo item(s).

It is also possible to deprioritize items as complete with an EXPRESSION using
the -e flag. Use -x to also process todo items that are normally invisible (as
with the 'ls' subcommand).\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urwid

from topydo.lib.Config import config


class ViewWidget(urwid.LineBox):
    def __init__(self, p_todolist):
        self._todolist = p_todolist

        self.titleedit = urwid.Edit("Title: ", "")
        self.sortedit = urwid.Edit("Sort expression: ", "")
        self.groupedit = urwid.Edit("Group expression: ", "")
        self.filteredit = urwid.Edit("Filter expression: ", "")

        radiogroup = []
        self.relevantradio = urwid.RadioButton(radiogroup, "Only show relevant todo items", True)
        self.allradio = urwid.RadioButton(radiogroup, "Show all todo items")

        self.pile = urwid.Pile([
            self.filteredit,
            self.titleedit,
            self.sortedit,
            self.groupedit,
            self.relevantradio,
            self.allradio,
            urwid.Button("Save", lambda _: urwid.emit_signal(self, 'save')),
            urwid.Button("Cancel", lambda _: self.close()),
        ])

        self.reset()

        super().__init__(self.pile)

        urwid.register_signal(ViewWidget, ['save', 'close'])

    @property
    def data(self):
        return {
            'title': self.titleedit.edit_text or self.filteredit.edit_text,
            'sortexpr': self.sortedit.edit_text or config().sort_string(),
            'groupexpr': self.groupedit.edit_text or config().group_string(),
            'filterexpr': self.filteredit.edit_text,
            'show_all': self.allradio.state,
        }

    @data.setter
    def data(self, p_data):
        self.titleedit.edit_text = p_data['title']
        self.sortedit.edit_text = p_data['sortexpr']
        self.groupedit.edit_text = p_data['groupexpr']
        self.filteredit.edit_text = p_data['filterexpr']
        self.relevantradio.set_state(not p_data['show_all'])
        self.allradio.set_state(p_data['show_all'])

    def reset(self):
        """ Resets the form. """
        self.titleedit.set_edit_text("")
        self.sortedit.set_edit_text("")
        self.filteredit.set_edit_text("")
        self.relevantradio.set_state(True)
        self.pile.focus_item = 0

    def close(self):
        urwid.emit_signal(self, 'close')

    def keypress(self, p_size, p_key):
        if p_key == 'esc':
            self.close()
        else:
            return super().keypress(p_size, p_key)  # pylint: disable=E1102



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import date

from topydo.lib.DCommand import DCommand
from topydo.lib.printers.PrettyPrinter import PrettyPrinter
from topydo.lib.Recurrence import NoRecurrenceException, advance_recurring_todo
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.lib.Utils import date_string_to_date


class DoCommand(DCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):

        self.strict_recurrence = False
        self.completion_date = date.today()

        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        self.condition = lambda todo: not todo.is_completed()
        self.condition_failed_text = "Todo has already been completed."

    def get_flags(self):
        """ Additional flags. """
        opts, long_opts = super().get_flags()

        return ("d:s" + opts, ["date=", "strict"] + long_opts)

    def process_flag(self, p_opt, p_value):
        super().process_flag(p_opt, p_value)

        if p_opt == "-s" or p_opt == "--strict":
            self.strict_recurrence = True
        elif p_opt == "-d" or p_opt == "--date":
            try:
                self.completion_date = relative_date_to_date(p_value)

                if not self.completion_date:
                    self.completion_date = date_string_to_date(p_value)
            except ValueError:
                self.completion_date = date.today()

    def _handle_recurrence(self, p_todo):
        if p_todo.has_tag('rec'):
            try:
                new_todo = advance_recurring_todo(
                    p_todo,
                    p_offset=self.completion_date,
                    p_strict=self.strict_recurrence
                )

                self.todolist.add_todo(new_todo)

            except NoRecurrenceException:
                self.error("Warning: todo item has an invalid recurrence pattern.")

    def prompt_text(self):
        return "Also mark subtasks as done? [y/N] "

    def prefix(self):
        return "Completed: "

    def execute_specific(self, p_todo):
        """ Actions specific to this command. """
        self._handle_recurrence(p_todo)
        self.execute_specific_core(p_todo)

        printer = PrettyPrinter()
        self.out(self.prefix() + printer.print_todo(p_todo))

    def execute_specific_core(self, p_todo):
        """
        The core operation on the todo itself. Also used to operate on
        child/parent tasks.
        """
        self.todolist.set_todo_completed(p_todo, self.completion_date)

    def usage(self):
        return """\
Synopsis: do [--date <DATE>] [--force] [--strict] <NUMBER 1> [<NUMBER 2> ...]
          do [-x] -e <EXPRESSION>\
"""

    def help(self):
        return """Marks the todo(s) with given NUMBER(s) as complete.

It is also possible to mark todo items as complete with an EXPRESSION using the
-e flag. Use -x to also process todo items that are normally invisible (as with
the 'ls' subcommand).

In case a todo has subitems (dependencies), a question is asked whether the
subitems should be marked as completed as well. When --force is given, no
interaction is required and the subitems are not marked completed.

In case a completed todo is recurring, a new todo will be added to the list,
while the given todo item is marked as complete. The new date is calculated
based on the todo item's due date. If the due date is in the past, today's date
is used to calculate the new recurrence date. Using --strict prevents this, and
then the actual due date of the todo item is used to calculate the new
recurrence date. Note that a future due date is always used as such to
calculate the new due date.

Use --date to set a custom completion date.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib import Filter
from topydo.lib.Command import Command, InvalidCommandArgument
from topydo.lib.Config import config
from topydo.lib.printers.Dot import DotPrinter
from topydo.lib.printers.PrettyPrinter import pretty_printer_factory
from topydo.lib.Sorter import Sorter
from topydo.lib.TodoListBase import InvalidTodoException
from topydo.lib.View import View


class DepCommand(Command):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

        try:
            self.subsubcommand = self.argument(0)
        except InvalidCommandArgument:
            self.subsubcommand = None

        self.printer = pretty_printer_factory(self.todolist)

    def _handle_add(self):
        for from_todo, to_todo in self._get_todos():
            self.todolist.add_dependency(from_todo, to_todo)

    def _handle_rm(self):
        for from_todo, to_todo in self._get_todos():
            self.todolist.remove_dependency(from_todo, to_todo)

    def _get_todos(self):
        result = []

        def get_parent_dependencies():
            child_todo = first_todo
            sibling_todo = second_todo

            return [(parent, child_todo) for parent in self.todolist.parents(sibling_todo)]

        def get_child_dependencies():
            parent_todo = first_todo
            sibling_todo = second_todo

            return [(parent_todo, child) for child in self.todolist.children(sibling_todo)]

        get_before_dependency = lambda: [(second_todo, first_todo)]
        get_after_dependency = lambda: [(first_todo, second_todo)]

        operators = {
            "after": get_after_dependency,
            "before": get_before_dependency,
            "child-of": get_child_dependencies,
            "childof": get_child_dependencies,
            "children-of": get_child_dependencies,
            "childrenof": get_child_dependencies,
            "parent-of": get_parent_dependencies,
            "parentof": get_parent_dependencies,
            "parents-of": get_parent_dependencies,
            "parentsof": get_parent_dependencies,
            "partof": get_before_dependency,
            "to": get_after_dependency,
        }

        try:
            first_todo_nr = self.argument(1)
            operator = self.argument(2)

            if operator in operators:
                second_todo_nr = self.argument(3)
            else:
                second_todo_nr = self.argument(2)
                operator = "to"

            first_todo = self.todolist.todo(first_todo_nr)
            second_todo = self.todolist.todo(second_todo_nr)

            result = operators[operator]()
        except (InvalidTodoException):
            self.error("Invalid todo number given.")
        except InvalidCommandArgument:
            self.error(self.usage())

        return result

    def _handle_ls(self):
        """ Handles the ls subsubcommand. """
        try:
            arg1 = self.argument(1)
            arg2 = self.argument(2)

            todos = []
            if arg2 == 'to' or arg1 == 'before':
                # dep ls 1 to OR dep ls before 1
                number = arg1 if arg2 == 'to' else arg2
                todo = self.todolist.todo(number)
                todos = self.todolist.children(todo)
            elif arg1 in {'to', 'after'}:
                # dep ls to 1 OR dep ls after 1
                number = arg2
                todo = self.todolist.todo(number)
                todos = self.todolist.parents(todo)
            else:
                raise InvalidCommandArgument

            sorter = Sorter(config().sort_string())
            instance_filter = Filter.InstanceFilter(todos)
            view = View(sorter, [instance_filter], self.todolist)
            self.out(self.printer.print_list(view.todos))
        except InvalidTodoException:
            self.error("Invalid todo number given.")
        except InvalidCommandArgument:
            self.error(self.usage())

    def _handle_dot(self):
        """ Handles the dot subsubcommand. """
        self.printer = DotPrinter(self.todolist)

        try:
            arg = self.argument(1)
            todo = self.todolist.todo(arg)
            arg = self.argument(1)
            todos = set([self.todolist.todo(arg)])
            todos |= set(self.todolist.children(todo))
            todos |= set(self.todolist.parents(todo))
            todos = sorted(todos, key=lambda t: t.text())

            self.out(self.printer.print_list(todos))
        except InvalidTodoException:
            self.error("Invalid todo number given.")
        except InvalidCommandArgument:
            self.error(self.usage())


    def execute(self):
        if not super().execute():
            return False

        dispatch = {
            'add':   self._handle_add,
            'rm':    self._handle_rm,
            'del':   self._handle_rm,
            'ls':    self._handle_ls,
            'clean': self.todolist.clean_dependencies,
            'dot':   self._handle_dot,
            'gc':    self.todolist.clean_dependencies,
        }

        if self.subsubcommand in dispatch:
            dispatch[self.subsubcommand]()
        else:
            self.error(self.usage())

    def usage(self):
        return """Synopsis:
  dep <add|rm> <NUMBER> [to] <NUMBER>
  dep add <NUMBER> <before|partof|after|parents-of|children-of> <NUMBER>
  dep ls <NUMBER> to
  dep ls to <NUMBER>
  dep ls <before|after> <NUMBER>
  dep dot <NUMBER>
  dep clean"""

    def help(self):
        return """\
* add               : Adds a dependency. `dep add 1 2` denotes that todo item 1
                      is dependant on todo item 2, i.e. item 2 is a subitem of
                      item 1.
* rm (alias: del)   : Removes a dependency.
* ls                : Lists all dependencies to or from a certain todo.
* dot               : Prints a dependency tree as a Dot graph.
* clean (alias: gc) : Removes redundant id or p tags.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This module provides functionality to sort lists with todo items. """

import re
from collections import OrderedDict, namedtuple
from datetime import date
from itertools import groupby

from topydo.lib.Config import config
from topydo.lib.Importance import average_importance, importance
from topydo.lib.Utils import date_string_to_date, humanize_date

Field = namedtuple('Field', ['sort', 'group', 'label'])

FIELDS = {
    'completed': Field(
        # when a task has no completion date, push it to the end by assigning it
        # the maximum possible date.
        sort=(lambda t: t.completion_date() if t.completion_date() else date.max),
        group=(lambda t: humanize_date(t.completion_date()) if t.completion_date() else 'None'),
        label='Completed',
    ),
    'context': Field(
        sort=lambda t: sorted(c.lower() for c in t.contexts()) or ['zz'],
        group=lambda t: sorted(t.contexts()) or ['None'],
        label='Context'
    ),
    'created': Field(
        # when a task has no creation date, push it to the end by assigning it
        # the maximum possible date.
        sort=(lambda t: t.creation_date() if t.creation_date() else date.max),
        group=(lambda t: humanize_date(t.creation_date()) if t.creation_date() else 'None'),
        label='Created',
    ),
    'importance': Field(
        sort=importance,
        group=importance,
        label='Importance',
    ),
    'importance-avg': Field(
        sort= average_importance,
        group=lambda t: round(average_importance(t), 1),
        label='Importance (avg)',
    ),
    'length': Field(
        sort=lambda t: t.length(),
        group=lambda t: t.length(),
        label='Length',
    ),
    'priority': Field(
        sort=(lambda t: t.priority() or 'M'),
        group=(lambda t: t.priority() or 'None'),
        label='Priority',
    ),
    'project': Field(
        sort=lambda t: sorted(p.lower() for p in t.projects()) or ['zz'],
        group=lambda t: sorted(t.projects()) or ['None'],
        label='Project',
    ),
    'text': Field(
        sort=lambda t: t.text().lower(),
        group=lambda t: t.text(),
        label='Text',
    ),
}

# map UI properties to properties in the FIELDS hash
FIELD_MAP = {
    'completed': 'completed',
    'completion': 'completed',
    'completion_date': 'completed',
    'done': 'completed',

    'context': 'context',
    'contexts': 'context',

    'created': 'created',
    'creation': 'created',
    'creation_date': 'created',

    'importance': 'importance',

    'importance-avg': 'importance-avg',
    'importance-average': 'importance-avg',

    'length': 'length',
    'len': 'length',

    'prio': 'priority',
    'priorities': 'priority',
    'priority': 'priority',

    'project': 'project',
    'projects': 'project',

    'text': 'text',
}

def _apply_sort_functions(p_todos, p_functions):
    sorted_todos = p_todos

    for function, order in reversed(p_functions):
        sorted_todos = sorted(sorted_todos, key=function,
                              reverse=(order == 'desc'))

    return sorted_todos


class Sorter(object):
    """
    This class sorts a todo list.

    Upon instantiation, a sort string should be passed to the class. Then, a
    list of todos must be passed to the sort method, which returns a copy of
    the list according to the sort string.

    A sort string is a comma separated line of field names, possibly prefixed
    with 'asc:' or 'desc' to denote the order in which that field must be
    sorted.

    Example:

        desc:importance,priority,asc:creation

    Meaning: a descending sort on the importance value, if equal, an ascending
    sort on priority and finally, if still equal an ascending sort on the
    creation field. Note that ascending is the default.

    The idea is that a list of sort functions is gathered, where the most
    specific search is done first. This relies on the fact that sorting is
    stable.
    """

    def __init__(self, p_sortstring="desc:priority", p_groupstring=""):
        def parse(p_string, p_group):
            """
            Parses a sort/group string and returns a list of functions and the
            desired order.
            """
            def get_field_function(p_field, p_group=False):
                """
                Turns a field, part of a sort/group string, into a lambda that
                takes a todo item and returns the field value.
                """
                compose = lambda i: i.sort if not p_group else (i.group, i.label)

                def group_value(p_todo):
                    """
                    Returns a value to assign the given todo to a group. Date tags
                    are grouped according to the relative date (1 day, 1 month,
                    ...)
                    """
                    result = 'No value'

                    if p_todo.has_tag(p_field):
                        if p_field == config().tag_due():
                            result = humanize_date(p_todo.due_date())
                        elif p_field == config().tag_start():
                            result = humanize_date(p_todo.start_date())
                        else:
                            result = p_todo.tag_value(p_field)

                            try:
                                result = humanize_date(date_string_to_date(result))
                            except ValueError:
                                pass

                    return result

                if p_field in FIELD_MAP:
                    return compose(FIELDS[FIELD_MAP[p_field]])
                else:
                    # treat it as a tag value
                    return compose(Field(
                        sort=lambda t: '0' + t.tag_value(p_field) if t.has_tag(p_field) else '1',
                        group=group_value,
                        label=p_field,
                    ))

            result = []
            fields = p_string.lower().split(',')

            for field in fields:
                parsed_field = re.match(
                    r'(?P<order>(asc|desc)(ending)?:)?(?P<field>\S+)',
                    field)

                if not parsed_field:
                    continue

                order = parsed_field.group('order')
                order = 'desc' if order and order.startswith('desc') else 'asc'

                field = parsed_field.group('field')
                if field:
                    function = get_field_function(field, p_group)

                    # reverse order for priority: lower characters have higher
                    # priority
                    if field in FIELD_MAP and FIELD_MAP[field] == 'priority':
                        order = 'asc' if order == 'desc' else 'desc'

                    result.append((function, order))

            return result

        self.groupfunctions = parse(p_groupstring, p_group=True) if p_groupstring else []
        self.pregroupfunctions = parse(p_groupstring, p_group=False) if p_groupstring else []
        self.sortfunctions = parse(p_sortstring, p_group=False)

    def sort(self, p_todos):
        """
        Sorts the list of todos given as a parameter, returns a new sorted
        list.

        The list is traversed in reverse order, such that the most specific
        sort operation is done first, relying on the stability of the sorted()
        function.
        """
        return _apply_sort_functions(p_todos, self.sortfunctions)

    def group(self, p_todos):
        """
        Groups the todos according to the given group string.
        """
        # preorder todos for the group sort
        p_todos = _apply_sort_functions(p_todos, self.pregroupfunctions)

        # initialize result with a single group
        result = OrderedDict([((), p_todos)])

        for (function, label), _ in self.groupfunctions:
            oldresult = result
            result = OrderedDict()
            for oldkey, oldgroup in oldresult.items():
                for key, _group in groupby(oldgroup, function):
                    newgroup = list(_group)

                    if not isinstance(key, list):
                        key = [key]

                    for subkey in key:
                        subkey = "{}: {}".format(label, subkey)
                        newkey = oldkey + (subkey,)

                        if newkey in result:
                            result[newkey] = result[newkey] + newgroup
                        else:
                            result[newkey] = newgroup

        # sort all groups
        for key, _group in result.items():
            result[key] = self.sort(_group)

        return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from configparser import NoOptionError, RawConfigParser
from os.path import expanduser

from topydo.lib.Config import config, home_config_path


def columns(p_alt_layout_path=None):
    """
    Returns list with complete column configuration dicts.
    """
    def _get_column_dict(p_cp, p_column):
        column_dict = dict()

        filterexpr = p_cp.get(p_column, 'filterexpr')

        try:
            title = p_cp.get(p_column, 'title')
        except NoOptionError:
            title = filterexpr

        column_dict['title'] = title or 'Yet another column'
        column_dict['filterexpr'] = filterexpr
        column_dict['sortexpr'] = p_cp.get(p_column, 'sortexpr')
        column_dict['groupexpr'] = p_cp.get(p_column, 'groupexpr')
        column_dict['show_all'] = p_cp.getboolean(p_column, 'show_all')

        return column_dict

    defaults = {
            'filterexpr': '',
            'sortexpr': config().sort_string(),
            'groupexpr': config().group_string(),
            'show_all': '0',
    }

    cp = RawConfigParser(defaults, strict=False)
    files = [
        "topydo_columns.ini",
        "topydo_columns.conf",
        ".topydo_columns",
        home_config_path('.topydo_columns'),
        home_config_path('.config/topydo/columns'),
        "/etc/topydo_columns.conf",
    ]

    if p_alt_layout_path is not None:
        files.insert(0, expanduser(p_alt_layout_path))
    for filename in files:
        if cp.read(filename):
            break

    column_list = []

    for column in cp.sections():
        column_list.append(_get_column_dict(cp, column))

    return column_list



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Provides the AddCommand class that implements the 'add' subcommand. """

import codecs
import re
from datetime import date
from os.path import expanduser
from sys import stdin

from topydo.lib.Config import config
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.WriteCommand import WriteCommand


class AddCommand(WriteCommand):
    def __init__(self, p_args, p_todolist, # pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)
        self.text = ' '.join(p_args)
        self.from_file = None

    def _process_flags(self):
        opts, args = self.getopt('f:')

        for opt, value in opts:
            if opt == '-f':
                self.from_file = expanduser(value)

        self.args = args

    def get_todos_from_file(self):
        if self.from_file == '-':
            f = stdin
        else:
            f = codecs.open(self.from_file, 'r', encoding='utf-8')

        todos = f.read().splitlines()

        return todos

    def _add_todo(self, p_todo_text):
        def _preprocess_input_todo(p_todo_text):
            """
            Pre-processes user input when adding a task.

            It detects a priority mid-sentence and puts it at the start.
            """
            todo_text = re.sub(r'^(.+) (\([A-Z]\))(.*)$', r'\2 \1\3',
                               p_todo_text)

            return todo_text

        todo_text = _preprocess_input_todo(p_todo_text)
        todo = self.todolist.add(todo_text)
        self.postprocess_input_todo(todo)

        if config().auto_creation_date():
            todo.set_creation_date(date.today())

        self.out(self.printer.print_todo(todo))

    def execute(self):
        """ Adds a todo item to the list. """
        if not super().execute():
            return False

        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))
        self._process_flags()

        if self.from_file:
            try:
                new_todos = self.get_todos_from_file()

                for todo in new_todos:
                    self._add_todo(todo)
            except (IOError, OSError):
                self.error('File not found: ' + self.from_file)
        else:
            if self.text:
                self._add_todo(self.text)
            else:
                self.error(self.usage())

    def usage(self):
        return """Synopsis:
  add <TEXT>
  add -f <FILE> | -"""

    def help(self):
        return """\
This subcommand automatically adds the creation date to the added item.

TEXT may contain:

* Priorities mid-sentence. Example: add "Water flowers (C)"

* Dependencies using before, after, partof, parents-of and children-of tags.
  These are translated to the corresponding 'id' and 'p' tags. The values of
  these tags correspond to the todo number (not the dependency number).

  Example: add "Subtask partof:1"

-f : Add todo items from specified FILE or from standard input.\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Command import Command


class ListProjectCommand(Command):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(
            p_args, p_todolist, p_out, p_err, p_prompt)

    def execute(self):
        if not super().execute():
            return False

        for project in sorted(self.todolist.projects(), key=lambda s: s.lower()):
            self.out(project)

    def usage(self):
        return """Synopsis: lsprj"""

    def help(self):
        return """Lists all projects in the todo list."""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2016 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class PaletteItem:
    PROJECT = 'project'
    PROJECT_FOCUS = 'project_focus'
    CONTEXT = 'context'
    CONTEXT_FOCUS = 'context_focus'
    METADATA = 'metadata'
    METADATA_FOCUS = 'metadata_focus'
    LINK = 'link'
    LINK_FOCUS = 'link_focus'

    DEFAULT = 'default'
    DEFAULT_FOCUS = 'default_focus'
    MARKED = 'marked'


def to_urwid_color(p_color):
    """
    Given a Color object, transform it to a color that urwid understands.
    """
    if not p_color.is_valid():
        return 'black'
    elif p_color.is_neutral():
        return 'default'
    else:
        return 'h{}'.format(p_color.color)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides a completer class that can be used by the prompt provided
by the prompt toolkit.
"""

import re

from prompt_toolkit.completion import Completer, Completion

from topydo.lib.Config import config
from topydo.lib.RelativeDate import relative_date_to_date
from topydo.ui.CompleterBase import CompleterBase, date_suggestions


def _dates(p_word_before_cursor):
    """ Generator for date completion. """
    to_absolute = lambda s: relative_date_to_date(s).isoformat()

    start_value_pos = p_word_before_cursor.find(':') + 1
    value = p_word_before_cursor[start_value_pos:]

    for reldate in date_suggestions():
        if not reldate.startswith(value):
            continue

        yield Completion(reldate, -len(value), display_meta=to_absolute(reldate))


class PromptCompleter(CompleterBase, Completer):
    """
    Completer class that completes projects, contexts, dates and
    subcommands and is compatible with prompt toolkit.
    """

    def _completion_generator(self, p_word, is_first_word):
        candidates = super().get_completions(p_word, is_first_word)
        for candidate in candidates:
            yield Completion(candidate, -len(p_word))

    def get_completions(self, p_word, p_is_first_word=False):
        # include all characters except whitespaces (for + and @)
        word_before_cursor = p_word.get_word_before_cursor(True)
        is_first_word = not re.match(r'\s*\S+\s',
                                     p_word.current_line_before_cursor)
        if word_before_cursor.startswith(config().tag_due() + ':'):
            return _dates(word_before_cursor)
        elif word_before_cursor.startswith(config().tag_start() + ':'):
            return _dates(word_before_cursor)

        return self._completion_generator(word_before_cursor, is_first_word)



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from topydo.lib.Config import config
from topydo.lib.prettyprinters.Numbers import PrettyPrinterNumbers
from topydo.lib.TodoListBase import InvalidTodoException
from topydo.lib.TodoParser import parse_line
from topydo.lib.WriteCommand import WriteCommand


class ModifyCommand(WriteCommand):
    def __init__(self, p_args, p_todolist, #pragma: no branch
                 p_out=lambda a: None,
                 p_err=lambda a: None,
                 p_prompt=lambda a: None):
        super().__init__(p_args, p_todolist, p_out, p_err,
                p_prompt)

    def execute(self):
        if not super().execute():
            return False

        if len(self.args) < 2:
            self.error(self.usage())
            return False
        text = self.args[0]
        numbers = self.args[1:]

        if not isinstance(text, str):
            self.error(self.usage())
            return False
        new_text_parsed = parse_line(text)
        new_tags = new_text_parsed['tags']

        self.printer.add_filter(PrettyPrinterNumbers(self.todolist))

        # Parse numbers/ids first to ensure all valid before any writes
        todos = []
        for num in numbers:
            try:
                todos.append(self.todolist.todo(num))
            except InvalidTodoException:
                self.error(f"Invalid todo number: {num}")
                return False

        for todo in todos:
            # Remove any existing start or due tags
            for tag in new_tags:
                if tag in (config().tag_start(), config().tag_due()):
                    todo.remove_tag(tag)

            self.todolist.append(todo, text)
            self.postprocess_input_todo(todo)

            self.out(self.printer.print_todo(todo))

        return True

    def usage(self):
        return """Synopsis: modify <TEXT> <NUMBER> [<NUMBER 2> ...]"""

    def help(self):
        return """\
Adds the TEXT to the end of the todo items indicated by NUMBER(s).\
"""



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Entry file for the Python todo.txt CLI. """

import sys

from topydo.Commands import get_subcommand
from topydo.lib import TodoFile, TodoList
from topydo.lib.Config import ConfigError, config
from topydo.ui.CLIApplicationBase import CLIApplicationBase, error

# First thing is to poke the configuration and check whether it's sane
# The modules below may already read in configuration upon import, so
# make sure to bail out if the configuration is invalid.
try:
    config()
except ConfigError as config_error:
    error(str(config_error))
    sys.exit(1)



class CLIApplication(CLIApplicationBase):
    """
    Class that represents the (original) Command Line Interface of Topydo.
    """

    def __init__(self):
        super().__init__()

    def run(self):
        """ Main entry function. """
        args = self._process_flags()

        self.todofile = TodoFile.TodoFile(config().todotxt())
        self.todolist = TodoList.TodoList(self.todofile.read())

        try:
            (subcommand, args) = get_subcommand(args)
        except ConfigError as ce:
            error('Error: ' + str(ce) + '. Check your aliases configuration')
            sys.exit(1)

        if subcommand is None:
            CLIApplicationBase._usage()

        if self._execute(subcommand, args) == False:
            sys.exit(1)
        else:
            self._post_execute()


def main():
    """ Main entry point of the CLI. """
    CLIApplication().run()

if __name__ == '__main__':
    main()
